"""Some functions are extracted from [ase](https://wiki.fysik.dtu.dk/ase/index.html)."""

from typing import List

import numpy as np
from loguru import logger

Na = 6.02214179e23  # Avogadro constant unit /mol
h = 6.6260696e-34  # Planck constant Unit J*s
kB = 1.3806503e-23  # Boltzmann constant J/K
R = Na * kB  # Ideal gas constant J/(K*mol)
amu = 1.66053906660e-27  # atomic mass unit kg
k = 1.380649e-23 / 1.602176634e-19  # eV/K
atomic_masses_iupac2016 = np.asarray(
    [
        1.0,  # X
        1.008,  # H [1.00784, 1.00811]
        4.002602,  # He
        6.94,  # Li [6.938, 6.997]
        9.0121831,  # Be
        10.81,  # B [10.806, 10.821]
        12.011,  # C [12.0096, 12.0116]
        14.007,  # N [14.00643, 14.00728]
        15.999,  # O [15.99903, 15.99977]
        18.998403163,  # F
        20.1797,  # Ne
        22.98976928,  # Na
        24.305,  # Mg [24.304, 24.307]
        26.9815385,  # Al
        28.085,  # Si [28.084, 28.086]
        30.973761998,  # P
        32.06,  # S [32.059, 32.076]
        35.45,  # Cl [35.446, 35.457]
        39.948,  # Ar
        39.0983,  # K
        40.078,  # Ca
        44.955908,  # Sc
        47.867,  # Ti
        50.9415,  # V
        51.9961,  # Cr
        54.938044,  # Mn
        55.845,  # Fe
        58.933194,  # Co
        58.6934,  # Ni
        63.546,  # Cu
        65.38,  # Zn
        69.723,  # Ga
        72.630,  # Ge
        74.921595,  # As
        78.971,  # Se
        79.904,  # Br [79.901, 79.907]
        83.798,  # Kr
        85.4678,  # Rb
        87.62,  # Sr
        88.90584,  # Y
        91.224,  # Zr
        92.90637,  # Nb
        95.95,  # Mo
        97.90721,  # 98Tc
        101.07,  # Ru
        102.90550,  # Rh
        106.42,  # Pd
        107.8682,  # Ag
        112.414,  # Cd
        114.818,  # In
        118.710,  # Sn
        121.760,  # Sb
        127.60,  # Te
        126.90447,  # I
        131.293,  # Xe
        132.90545196,  # Cs
        137.327,  # Ba
        138.90547,  # La
        140.116,  # Ce
        140.90766,  # Pr
        144.242,  # Nd
        144.91276,  # 145Pm
        150.36,  # Sm
        151.964,  # Eu
        157.25,  # Gd
        158.92535,  # Tb
        162.500,  # Dy
        164.93033,  # Ho
        167.259,  # Er
        168.93422,  # Tm
        173.054,  # Yb
        174.9668,  # Lu
        178.49,  # Hf
        180.94788,  # Ta
        183.84,  # W
        186.207,  # Re
        190.23,  # Os
        192.217,  # Ir
        195.084,  # Pt
        196.966569,  # Au
        200.592,  # Hg
        204.38,  # Tl [204.382, 204.385]
        207.2,  # Pb
        208.98040,  # Bi
        208.98243,  # 209Po
        209.98715,  # 210At
        222.01758,  # 222Rn
        223.01974,  # 223Fr
        226.02541,  # 226Ra
        227.02775,  # 227Ac
        232.0377,  # Th
        231.03588,  # Pa
        238.02891,  # U
        237.04817,  # 237Np
        244.06421,  # 244Pu
        243.06138,  # 243Am
        247.07035,  # 247Cm
        247.07031,  # 247Bk
        251.07959,  # 251Cf
        252.0830,  # 252Es
        257.09511,  # 257Fm
        258.09843,  # 258Md
        259.1010,  # 259No
        262.110,  # 262Lr
        267.122,  # 267Rf
        268.126,  # 268Db
        271.134,  # 271Sg
        270.133,  # 270Bh
        269.1338,  # 269Hs
        278.156,  # 278Mt
        281.165,  # 281Ds
        281.166,  # 281Rg
        285.177,  # 285Cn
        286.182,  # 286Nh
        289.190,  # 289Fl
        289.194,  # 289Mc
        293.204,  # 293Lv
        293.208,  # 293Ts
        294.214,  # 294Og
    ],
)

chemical_symbols = [
    # 0
    "X",
    # 1
    "H",
    "He",
    # 2
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    # 3
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    # 4
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    # 5
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    # 6
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    # 7
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
]

atomic_numbers = {}
for Z, symbol in enumerate(chemical_symbols):
    atomic_numbers[symbol] = Z


def elements2masses(elements: List[str]) -> np.ndarray:
    """Convert a list of elements to a list of masses

    Parameters
    ----------
    elements:
        Elements list

    Returns
    -------
    List[float]
        List[Quality]

    Examples
    --------
    >>> from dspawpy.io.utils import elements2masses
    >>> elements = ["H", "O"]
    >>> masses = elements2masses(elements)
    >>> masses
    [1.008, 15.999]

    """
    masses = []
    for e in elements:
        masses.append(atomic_masses_iupac2016[atomic_numbers[e]])
    return np.asarray(masses).tolist()


def get_ma(elements, positions, Natom):
    """Get the moments of inertia along the principal axes.

    The three principal moments of inertia are computed from the
    eigenvalues of the symmetric inertial tensor. Periodic boundary
    conditions are ignored. Units of the moments of inertia are
    amu*angstrom**2.
    """
    masses = elements2masses(elements)
    com = np.dot(masses, positions) / np.sum(masses)
    positions -= com  # translate center of mass to origin
    masses = elements2masses(elements)

    # Initialize elements of the inertial tensor
    I11 = I22 = I33 = I12 = I13 = I23 = 0.0
    for i in range(Natom):
        x, y, z = positions[i]
        m = masses[i]

        I11 += m * (y**2 + z**2)
        I22 += m * (x**2 + z**2)
        I33 += m * (x**2 + y**2)
        I12 += -m * x * y
        I13 += -m * x * z
        I23 += -m * y * z

    inertia = np.asarray([[I11, I12, I13], [I12, I22, I23], [I13, I23, I33]])

    evals, evecs = np.linalg.eigh(inertia)
    return evals


class IdealGasThermo:
    """import from ase.thermochemistry.IdealGasThermo

    Parameters
    ----------
    vib_energies:
        List of vibrational energies in eV.
    geometry:
        One of 'linear', 'nonlinear', 'monatomic'
    potentialenergy:
        Potential energy in eV.
    elements:
        such as ['H', 'O'].
    symmetrynumber:
        Symmetry number.
    spin:
        Spin multiplicity.
    natoms:
        Number of atoms.

    Examples
    --------
    >>> from dspawpy.io.utils import IdealGasThermo
    >>> thermo = IdealGasThermo(vib_energies=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6],
    ...                         geometry='linear', potentialenergy=0.,  # eV
    ...                         elements=['H', 'O'], positions=[[0, 0, 0], [0, 0, 1]],  # angstrom
    ...                         symmetrynumber=None, spin=None, natoms=None)
    >>> print(thermo.get_enthalpy(298.15))  # K
    Enthalpy components at T = 298.15 K:
    ===============================
    E_pot                  0.000 eV
    E_ZPE                  0.300 eV
    Cv_trans (0->T)        0.039 eV
    Cv_rot (0->T)          0.026 eV
    Cv_vib (0->T)          0.000 eV
    (C_v -> C_p)           0.026 eV
    -------------------------------
    H                      0.390 eV
    ===============================
    0.389924026967057

    """

    def __init__(
        self,
        vib_energies,
        geometry,
        potentialenergy=0.0,
        elements=None,
        positions=None,
        symmetrynumber=None,
        spin=None,
        natoms=None,
    ):
        self.potentialenergy = potentialenergy
        self.geometry = geometry
        self.elements = elements
        if isinstance(positions, list):
            self.positions = np.asarray(positions, dtype=float)
        elif isinstance(positions, np.ndarray):
            self.positions = positions
        else:
            raise TypeError("positions must be list or np.ndarray")
        if isinstance(vib_energies, list):
            vib_energies = np.asarray(vib_energies)
        elif isinstance(vib_energies, np.ndarray):
            pass
        else:
            raise TypeError("vib_energies must be list or np.ndarray")
        self.sigma = symmetrynumber
        self.spin = spin
        if natoms is None:
            if elements:
                natoms = len(elements)
        # Cut the vibrations to those needed from the geometry.
        if natoms:
            if geometry == "nonlinear":
                self.vib_energies = vib_energies[-(3 * natoms - 6) :]
            elif geometry == "linear":
                self.vib_energies = vib_energies[-(3 * natoms - 5) :]
            elif geometry == "monatomic":
                self.vib_energies = []
        else:
            self.vib_energies = vib_energies
        # Make sure no imaginary frequencies remain.
        if sum(np.iscomplex(self.vib_energies)):
            raise ValueError("Imaginary frequencies are present.")
        else:
            self.vib_energies = np.real(self.vib_energies)  # clear +0.j
        self.referencepressure = 1.0e5  # Pa
        self.natoms = natoms

    def get_ZPE_correction(self):
        """Returns the zero-point vibrational energy correction in eV."""
        zpe = 0.0
        for energy in self.vib_energies:
            zpe += 0.5 * energy
        return zpe

    def _vibrational_energy_contribution(self, temperature):
        """Calculates the change in internal energy due to vibrations from
        0K to the specified temperature for a set of vibrations given in
        eV and a temperature given in Kelvin. Returns the energy change
        in eV.
        """
        kT = k * temperature
        dU = 0.0
        for energy in self.vib_energies:
            dU += energy / (np.exp(energy / kT) - 1.0)
        return dU

    def _vibrational_entropy_contribution(self, temperature):
        """Calculates the entropy due to vibrations for a set of vibrations
        given in eV and a temperature given in Kelvin.  Returns the entropy
        in eV/K.
        """
        kT = k * temperature
        S_v = 0.0
        for energy in self.vib_energies:
            x = energy / kT
            S_v += x / (np.exp(x) - 1.0) - np.log(1.0 - np.exp(-x))
        S_v *= k
        return S_v

    def get_enthalpy(self, temperature):
        """Returns the enthalpy, in eV, in the ideal gas approximation
        at a specified temperature (K).
        """
        fmt = "%-15s%13.3f eV"
        print("Enthalpy components at T = %.2f K:" % temperature)
        print("=" * 31)

        H = 0.0

        print(fmt % ("E_pot", self.potentialenergy))
        H += self.potentialenergy

        zpe = self.get_ZPE_correction()
        print(fmt % ("E_ZPE", zpe))
        H += zpe

        Cv_t = 3.0 / 2.0 * k  # translational heat capacity (3-d gas)
        print(fmt % ("Cv_trans (0->T)", Cv_t * temperature))
        H += Cv_t * temperature

        if self.geometry == "nonlinear":  # rotational heat capacity
            Cv_r = 3.0 / 2.0 * k
        elif self.geometry == "linear":
            Cv_r = k
        elif self.geometry == "monatomic":
            Cv_r = 0.0
        else:
            raise ValueError("Unknown geometry.")
        print(fmt % ("Cv_rot (0->T)", Cv_r * temperature))
        H += Cv_r * temperature

        dH_v = self._vibrational_energy_contribution(temperature)
        print(fmt % ("Cv_vib (0->T)", dH_v))
        H += dH_v

        Cp_corr = k * temperature
        print(fmt % ("(C_v -> C_p)", Cp_corr))
        H += Cp_corr

        print("-" * 31)
        print(fmt % ("H", H))
        print("=" * 31)
        return H

    def get_entropy(self, temperature, pressure):
        """Returns the entropy, in eV/K, in the ideal gas approximation
        at a specified temperature (K) and pressure (Pa).
        """
        if self.elements is None or self.sigma is None or self.spin is None:
            raise RuntimeError(
                "elements, symmetrynumber, and spin must be "
                "specified for entropy and free energy "
                "calculations.",
            )
        S = 0.0
        # Translational entropy (term inside the log is in SI units).
        mass = sum(elements2masses(self.elements)) * amu  # kg/molecule
        S_t = (2 * np.pi * mass * kB * temperature / h**2) ** (3.0 / 2)
        S_t *= kB * temperature / self.referencepressure
        S_t = k * (np.log(S_t) + 5.0 / 2.0)
        S += S_t

        # Rotational entropy (term inside the log is in SI units).
        if self.geometry == "monatomic":
            S_r = 0.0
        elif self.geometry == "nonlinear":
            inertias = (
                get_ma(self.elements, self.positions, self.natoms)
                * amu
                / (10.0**10) ** 2
            )  # kg m^2
            S_r = np.sqrt(np.pi * np.prod(inertias)) / self.sigma
            S_r *= (8.0 * np.pi**2 * kB * temperature / h**2) ** (3.0 / 2.0)
            S_r = k * (np.log(S_r) + 3.0 / 2.0)
        elif self.geometry == "linear":
            inertias = (
                get_ma(self.elements, self.positions, self.natoms)
                * amu
                / (10.0**10) ** 2
            )  # kg m^2
            inertia = max(inertias)  # should be two identical and one zero
            S_r = 8 * np.pi**2 * inertia * kB * temperature / self.sigma / h**2
            S_r = k * (np.log(S_r) + 1.0)
        else:
            raise ValueError("Unknown geometry.")
        S += S_r
        # Electronic entropy.
        S_e = k * np.log(2 * self.spin + 1)
        S += S_e
        # Vibrational entropy.
        S_v = self._vibrational_entropy_contribution(temperature)
        S += S_v
        # Pressure correction to translational entropy.
        S_p = -k * np.log(pressure / self.referencepressure)
        S += S_p
        return S

    def get_gibbs_energy(self, temperature, pressure):
        """Returns the Gibbs free energy, in eV, in the ideal gas
        approximation at a specified temperature (K) and pressure (Pa).
        """
        H = self.get_enthalpy(temperature)
        print("")
        S = self.get_entropy(temperature, pressure)
        G = H - temperature * S

        print("")
        print(
            "Free energy components at T = %.2f K and P = %.1f Pa:"
            % (temperature, pressure),
        )
        print("=" * 23)
        fmt = "%5s%15.3f eV"
        print(fmt % ("H", H))
        print(fmt % ("-T*S", -temperature * S))
        print("-" * 23)
        print(fmt % ("G", G))
        print("=" * 23)
        return G


def getTSgas(
    fretxt="frequency.txt",
    datafile=".",
    potentialenergy: float = 0.0,  # eV
    elements=None,
    geometry="linear",
    positions=None,  # Angstrom
    symmetrynumber=1,
    spin=1,
    temperature=298.15,
    pressure: float = 101325,
    verbose: bool = False,
):
    """Energy contribution to entropy under the ideal gas approximation"

    Parameters
    ----------
    fretxt:
        Path to the file recording frequency information, default is 'frequency.txt' in the current path
    datafile:
        Path to the JSON or h5 file or folder containing them, default to the current path;
        If set to None, the elements and positions parameters must be provided
    potentialenergy:
        Potential energy, unit eV
    elements:
        List of elements, if
    geometry:
        Molecular geometry, monatomic, linear, nonlinear
    positions:
        Atomic coordinates, unit Angstrom
    symmetrynumber:
        Symmetry number
    spin:
        Spin number
    temperature:
        Temperature, unit K
    pressure:
        Pressure, unit Pa

    Returns
    -------
    TSgas:
        Under the ideal gas approximation, calculates the energy contribution to entropy, in units of eV

    Examples
    --------
    >>> from dspawpy.io.utils import getTSgas
    >>> TSgas=getTSgas(fretxt='tests/2.13/frequency.txt', datafile='tests/2.13/frequency.h5', potentialenergy=-0.0,  geometry='linear', symmetrynumber=1, spin=1, temperature=298.15, pressure=101325.0)
    --> T*S (eV): 0.8515317035550232

    """
    import os

    abstxt = os.path.abspath(fretxt)
    ve = []
    with open(abstxt) as ft:
        lines = ft.readlines()
        for i in range(2, len(lines)):
            if lines[i].strip()[1] == "f/i":
                ve.append(complex(lines[i].split()[-1]) / 1000)
            else:
                ve.append(float(lines[i].split()[-1]) / 1000)
    if datafile is not None:
        absfile = get_absfile(datafile, task="frequency", verbose=verbose)
        if absfile.endswith(".h5"):
            from dspawpy.io.read import get_ele_from_h5

            elements = get_ele_from_h5(absfile)
            import h5py

            data = h5py.File(absfile)
            poses = np.asarray(data.get("/AtomInfo/Position")).reshape(-1, 3)

        elif absfile.endswith(".json"):
            import json

            with open(absfile) as f:
                data = json.load(f)
            atoms = data["AtomInfo"]["Atoms"]
            elements = []
            poses = []
            for i in range(len(atoms)):
                elements.append(atoms[i]["Element"])
                poses.append(atoms[i]["Position"])
        else:
            raise TypeError("Only support h5/json file")
    else:
        elements = elements
        poses = positions

    # Compute the energy contribution to entropy
    thermo = IdealGasThermo(
        vib_energies=ve,  # eV
        potentialenergy=potentialenergy,  # eV
        elements=elements,
        geometry=geometry,
        positions=poses,  # Angstrom
        symmetrynumber=symmetrynumber,
        spin=spin,
    )
    S = thermo.get_entropy(temperature, pressure)
    dE = S * temperature
    print(f"--> T*S (eV): {dE}")

    return S * temperature


def d_band(
    spin, dos_data
):  # Define the function, with two variables given in parentheses
    """Calculate the d-band center

    Parameters
    ----------
    spin : Spin.up or Spin.down
        Spin type,
    dos_data : pymatgen.electronic_structure.dos.CompleteDos
        Dos data

    Returns
    -------
    db1:
        d with center value

    Examples
    --------
    >>> from dspawpy.io.utils import d_band
    >>> from dspawpy.io.read import get_dos_data
    >>> dos_data = get_dos_data("tests/supplement/dos.h5")  # Reads data from dos.h5
    >>> for spin in dos_data.densities:
    ...     print('spin=', spin)
    ...     c = d_band(spin, dos_data)
    ...     print(c) # doctest: +ELLIPSIS
    spin= 1
    -3.666508...

    """
    from pymatgen.electronic_structure.core import OrbitalType

    dos_d = dos_data.get_spd_dos()[OrbitalType.d]
    # dos_d = dos_data.get_spd_dos()[d]
    Efermi = dos_data.efermi
    epsilon = dos_d.energies - Efermi  # shift d-band center

    N1 = dos_d.densities[spin]
    M1 = epsilon * N1
    from scipy import integrate

    SummaM1 = integrate.simpson(M1, x=epsilon)
    SummaN1 = integrate.simpson(N1, x=epsilon)

    return SummaM1 / SummaN1


def getZPE(fretxt: str = "frequency.txt"):
    """Read data from fretxt, calculate ZPE

    The results will also be saved to ZPE_TS.dat.

    Parameters
    ----------
    fretxt:
        Path to the file recording frequency information, default to 'frequency.txt' in the current path

    Returns
    -------
    ZPE:
        Zero-point energy

    Examples
    --------
    >>> from dspawpy.io.utils import getZPE
    >>> ZPE=getZPE(fretxt='tests/2.13/frequency.txt')
    --> Zero-point energy,  ZPE (eV): 0.1424200165

    """
    import os

    abstxt = os.path.abspath(fretxt)
    data_get_ZPE = []
    with open(abstxt) as f:
        for line in f.readlines():
            data_line = line.strip().split()
            if len(data_line) != 6:
                continue
            if data_line[1] == "f":
                data_get_ZPE.append(float(data_line[5]))

    data_get_ZPE = np.asarray(data_get_ZPE)

    assert len(data_get_ZPE) > 0, (
        "Only imaginary frequencies, please consider re-optimizing the structure"
    )

    ZPE = 0
    for data in data_get_ZPE:
        ZPE += data / 2000.0
    print("--> Zero-point energy,  ZPE (eV):", ZPE)

    return ZPE


def getTSads(
    fretxt: str = "frequency.txt",
    T: float = 298.15,
):
    """Read data from fretxt, calculate ZPE and TS

    Will also save the results to TSads.dat

    Parameters
    ----------
    fretxt:
        Path to the file recording frequency information, default 'frequency.txt' in the current path
    T:
        Temperature, unit K, default 298.15

    Returns
    -------
    TS:
        Entropy correction

    Examples
    --------
    >>> from dspawpy.io.utils import getTSads
    >>> TSads=getTSads(fretxt='tests/2.13/frequency.txt', T=298.15)
    --> T*S (eV): 4.7566997225177686e-06

    """
    import os

    abstxt = os.path.abspath(fretxt)

    data_get_TS = []
    with open(abstxt) as f:
        for line in f.readlines():
            data_line = line.strip().split()
            if len(data_line) != 6:
                continue
            if data_line[1] == "f":
                data_get_TS.append(float(data_line[2]))

    data_get_TS = np.asarray(data_get_TS)

    assert len(data_get_TS) > 0, (
        "Only imaginary frequencies, please consider re-optimizing the structure"
    )

    sum_S = 0
    import math  # because it will use e raised to a power, and the ln() logarithm

    for vi_THz in data_get_TS:
        vi_Hz = vi_THz * 1e12
        m1 = h * Na * vi_Hz
        m2 = h * vi_Hz / (kB * T)
        m3 = math.exp(m2) - 1
        m4 = T * m3
        m5 = 1 - math.exp(-m2)  # math.exp(3) is e raised to the power of 3
        m6 = math.log(
            m5, math.e
        )  # m6= ln(m5)   math.e in Python is e, logarithm with base on the right side
        m7 = R * m6
        m8 = m1 / m4 - m7  # S unit J/(mol*K)
        m9 = (
            T * m8 / 1000
        ) / 96.49  # T*S, converting units to kJ/mol, 96.49 kJ/mol = 1 eV unit eV
        sum_S += m9

    print("--> T*S (eV):", sum_S)

    return sum_S


def get_absfile(
    datafile: str,
    task: str,
    only_h5: bool = False,
    verbose: bool = False,
) -> str:
    """Return the absolute path of the desired data file based on the given datafile

    Parameters
    ----------
    datafile:
        Path to the h5/json data file or its directory, can be a relative path
    task:
        Calculate task type, such as scf, optical, etc.
    only_h5:
        Whether to only look for h5 files, default is False

    Raises
    ------
    FileNotFoundError
        - The specified folder path does not contain the corresponding h5/json data files
        - The specified file path does not exist

    Returns
    -------
    absfile:
        The absolute path to the required data file

    """
    assert datafile is not None, "datafile is None"
    import os

    absfile = os.path.abspath(datafile)
    if only_h5:
        if os.path.isdir(absfile):  # specified datafile is actually a directory
            directory = absfile  # search datafile in the given directory
            absh5 = os.path.join(directory, f"{task}.h5")
            if os.path.exists(absh5):
                absfile = absh5
            else:
                raise FileNotFoundError(f"No {absh5}/{absfile}")
        elif not os.path.isfile(absfile):
            raise FileNotFoundError(f"No {absfile}")
        else:
            assert absfile.endswith(
                ".h5",
            ), f"Only support h5 file, but got {absfile}"
    elif os.path.isdir(absfile):  # specified datafile is actually a directory
        directory = absfile  # search datafile in the given directory
        absh5 = os.path.join(directory, f"{task}.h5")
        absjs = os.path.join(directory, f"{task}.json")
        if os.path.exists(absh5):
            absfile = absh5
        elif os.path.exists(absjs):
            absfile = absjs
        else:
            raise FileNotFoundError(f"No {absh5}/{absjs}")
    elif not os.path.isfile(absfile):
        raise FileNotFoundError(f"No {absfile}")
    else:
        assert absfile.endswith(".h5") or absfile.endswith(
            ".json",
        ), f"Only support h5/json file, but got {absfile}"

    if verbose:
        logger.info(f"Reading {absfile}...")
    return absfile


# extract from ASE


def string2symbols(s: str) -> List[str]:
    """Convert string to list of chemical symbols."""
    return list(Formula(s))


def _symbols2numbers(symbols) -> List[int]:
    if isinstance(symbols, str):
        symbols = string2symbols(symbols)
    numbers = []
    for s in symbols:
        if isinstance(s, str):
            numbers.append(atomic_numbers[s])
        else:
            numbers.append(int(s))
    return numbers


class Formula:
    def __init__(
        self,
        formula: str = "",
        *,
        strict: bool = False,
        fmt: str = "",
        _tree=None,
        _count=None,
    ):
        """Chemical formula object.

        Parameters
        ----------
        formula:
            Text string representation of formula.  Examples: ``'6CO2'``,
            ``'30Cu+2CO'``, ``'Pt(CO)6'``.
        strict:
            Only allow real chemical symbols.
        fmt:
            Reorder according to *fmt*.  Must be one of hill, metal,
            abc or reduce.

        Examples
        --------
        >>> from dspawpy.io.utils import Formula
        >>> w = Formula('H2O')

        Raises
        ------
        ValueError
            on malformed formula

        """
        if fmt:
            if fmt not in {"hill", "metal", "abc", "reduce"}:
                raise ValueError(f"Illegal fmt: {fmt}")
            formula = format(Formula(formula), fmt)
        self._formula = formula
        self._tree = _tree or parse(formula)
        self._count = _count or count_tree(self._tree)
        if strict:
            for symbol in self._count:
                if symbol not in atomic_numbers:
                    raise ValueError("Unknown chemical symbol: " + symbol)

    def __iter__(self, tree=None):
        if tree is None:
            tree = self._tree
        if isinstance(tree, str):
            yield tree
        elif isinstance(tree, tuple):
            tree, N = tree
            for _ in range(N):
                yield from self.__iter__(tree)
        else:
            for tree in tree:
                yield from self.__iter__(tree)


def parse(f: str):  # -> Tree
    if not f:
        return []
    parts = f.split("+")
    result = []
    for part in parts:
        n, f = strip_number(part)
        result.append((parse2(f), n))
    return result


def strip_number(s: str):
    import re

    m = re.match("[0-9]*", s)
    assert m is not None
    return int(m.group() or 1), s[m.end() :]


def parse2(f: str):
    units = []
    import re

    while f:
        if f[0] == "(":
            level = 0
            for i, c in enumerate(f[1:], 1):
                if c == "(":
                    level += 1
                elif c == ")":
                    if level == 0:
                        break
                    level -= 1
            else:
                raise ValueError
            f2 = f[1:i]
            n, f = strip_number(f[i + 1 :])
            unit = (parse2(f2), n)
        else:
            m = re.match("([A-Z][a-z]?)([0-9]*)", f)
            if m is None:
                raise ValueError
            symb = m.group(1)
            number = m.group(2)
            if number:
                unit = (symb, int(number))
            else:
                unit = symb
            f = f[m.end() :]
        units.append(unit)
    if len(units) == 1:
        return units[0]
    return units


def count_tree(tree):
    if isinstance(tree, str):
        return {tree: 1}
    if isinstance(tree, tuple):
        tree, N = tree
        return {symb: n * N for symb, n in count_tree(tree).items()}
    dct = {}
    for tree in tree:
        for symb, n in count_tree(tree).items():
            m = dct.get(symb, 0)
            dct[symb] = m + n
    return dct


import functools  # noqa: E402
from pathlib import PurePath  # noqa: E402


class iofunction:
    """Decorate func so it accepts either str or file.

    (Won't work on functions that return a generator.)
    """

    def __init__(self, mode):
        self.mode = mode

    def __call__(self, func):
        @functools.wraps(func)
        def iofunc(file, *args, **kwargs):
            openandclose = isinstance(file, (str, PurePath))
            fd = None
            try:
                if openandclose:
                    fd = open(str(file), self.mode)
                else:
                    fd = file
                obj = func(fd, *args, **kwargs)
                return obj
            finally:
                if openandclose and fd is not None:
                    # fd may be None if open() failed
                    fd.close()

        return iofunc


def reader(func):
    return iofunction("r")(func)


def label_to_symbol(label: str):
    if len(label) >= 2:
        test_symbol = label[0].upper() + label[1].lower()
        if test_symbol in chemical_symbols:
            return test_symbol
    test_symbol = label[0].upper()
    if test_symbol in chemical_symbols:
        return test_symbol
    else:
        raise KeyError(f"Could not parse species from label {label}.")



import os
import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_input, get_inputs, get_lims, save_figure_dpi300
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse, Dselect


def pre_ele_dos(language: Literal["EN", "CN"]):
    def imp():
        global DosPlotter, get_dos_data
        from pymatgen.electronic_structure.plotter import DosPlotter

        from dspawpy.io.read import get_dos_data

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["dos"], completer=pc)
    D["xlims"] = get_lims(Dparameter[language][9], language)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["shift"] = get_input(Dselect[language][3], ["y", "n"])
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "dos.png"

    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, DosPlotter, get_dos_data


def s5_1(language: Literal["EN", "CN"]) -> dict:
    """Total density of states"""

    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import DosPlotter as cli_DosPlotter

        from dspawpy.io.read import get_dos_data as cli_get_dos_data

        for D in list_ds:
            if D["menu"] == 51:
                dos_data = cli_get_dos_data(D["inf"])
                logger.debug("got data, initializing plotter...")
                if D["shift"] == "y":
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                else:
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                dos_plotter.add_dos("total dos", dos=dos_data)
                dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
                save_figure_dpi300(D["figure"])

        D = {}
    else:
        D, DosPlotter, get_dos_data = pre_ele_dos(language)
        D["menu"] = 51
        dos_data = get_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        if D["shift"] == "y":
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        else:
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        logger.debug("plotter initialized")
        dos_plotter.add_dos("total dos", dos=dos_data)
        dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D


def pre_ele_pdos(language: Literal["EN", "CN"]):
    def imp():
        global DosPlotter, get_dos_data, CompleteDos
        from pymatgen.electronic_structure.plotter import DosPlotter

        from dspawpy.io.read import get_dos_data

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["pdos"], completer=pc)
    D["xlims"] = get_lims(Dparameter[language][9], language)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["shift"] = get_input(Dselect[language][3], ["y", "n"])
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "pdos.png"

    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, DosPlotter, get_dos_data


def s5_2(language: Literal["EN", "CN"]) -> dict:
    "Project the density of states onto different orbitals"

    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import DosPlotter as cli_DosPlotter

        from dspawpy.io.read import get_dos_data as cli_get_dos_data

        for D in list_ds:
            if D["menu"] == 52:
                dos_data = cli_get_dos_data(D["inf"])
                if D["shift"] == "y":
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                else:
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                dos_plotter.add_dos_dict(dos_data.get_spd_dos())  # type: ignore
                dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
                save_figure_dpi300(D["figure"])
        D = {}
    else:
        D, DosPlotter, get_dos_data = pre_ele_pdos(language)
        D["menu"] = 52
        dos_data = get_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        if D["shift"] == "y":
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        else:
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        logger.debug("plotter initialized")
        dos_plotter.add_dos_dict(dos_data.get_spd_dos())  # type: ignore
        dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D


def s5_3(language: Literal["EN", "CN"]) -> dict:
    "Project the density of states onto different elements"
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import DosPlotter as cli_DosPlotter

        from dspawpy.io.read import get_dos_data as cli_get_dos_data

        for D in list_ds:
            if D["menu"] == 52:
                dos_data = cli_get_dos_data(D["inf"])
                if D["shift"] == "y":
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                else:
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                dos_plotter.add_dos_dict(dos_data.get_element_dos())  # type: ignore
                dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
                save_figure_dpi300(D["figure"])
        D = {}
    else:
        D, DosPlotter, get_dos_data = pre_ele_pdos(language)
        D["menu"] = 53
        dos_data = get_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        if D["shift"] == "y":
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        else:
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        logger.debug("plotter initialized")
        dos_plotter.add_dos_dict(dos_data.get_element_dos())  # type: ignore
        dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D


def s5_4(language: Literal["EN", "CN"]) -> dict:
    "Project the density of states onto different orbitals of different atoms"
    if auto_test_cli:
        from pymatgen.electronic_structure.core import Orbital as cli_Orbital
        from pymatgen.electronic_structure.plotter import DosPlotter as cli_DosPlotter

        from dspawpy.io.read import get_dos_data as cli_get_dos_data

        for D in list_ds:
            if D["menu"] == 52:
                dos_data = cli_get_dos_data(D["inf"])
                if D["shift"] == "y":
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                else:
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)

                for _n, _os, _e in zip(D["ns"], D["oss"], D["es"]):
                    for _orb in _os:
                        logger.info(f"atom-{_n} {_orb}")
                        dos_plotter.add_dos(
                            f"{_e}(atom-{_n}) {_orb}",  # label
                            dos_data.get_site_orbital_dos(  # type: ignore
                                dos_data.structure[int(_n)],  # type: ignore
                                getattr(cli_Orbital, _orb),
                            ),
                        )
                dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
                save_figure_dpi300(D["figure"])
        D = {}
    else:

        def imp():
            global Orbital
            from pymatgen.electronic_structure.core import Orbital

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D, DosPlotter, get_dos_data = pre_ele_pdos(language)
        D["menu"] = 54
        dos_data = get_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        if D["shift"] == "y":
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        else:
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        logger.debug("plotter initialized")
        logger.info(dos_data.structure)  # type: ignore
        sites = dos_data.structure.sites  # type: ignore
        numbers = [str(i) for i in range(len(sites))]

        D["ns"] = []
        D["es"] = []
        D["oss"] = []
        _e = None
        while True:
            _n = get_input(Dselect[language][6], numbers, allow_empty=True)
            if _n == "":
                break
            available_orbitals = ["s"]
            _e = sites[int(_n)].specie
            orbitals = _e.atomic_orbitals
            assert isinstance(orbitals, dict)
            for o in orbitals:
                if "p" in o:
                    available_orbitals.append("p")
                    available_orbitals.append("px")
                    available_orbitals.append("py")
                    available_orbitals.append("pz")
                elif "d" in o:
                    available_orbitals.append("d")
                    available_orbitals.append("dxy")
                    available_orbitals.append("dyz")
                    available_orbitals.append("dxz")
                    available_orbitals.append("dx2")
                    available_orbitals.append("dz2")
                elif "f" in o:
                    available_orbitals.append("f")
                    available_orbitals.append("f_3")
                    available_orbitals.append("f_2")
                    available_orbitals.append("f_1")
                    available_orbitals.append("f0")
                    available_orbitals.append("f1")
                    available_orbitals.append("f2")
                    available_orbitals.append("f3")
            unique_orbitals = list(set(available_orbitals))
            _os = get_inputs(Dselect[language][7], unique_orbitals)
            D["ns"].append(_n)
            D["oss"].append(_os)
            D["es"].append(_e)

        assert _e is not None
        import_thread.join()
        logger.info(Dresponse[language][16])
        for _n, _os, _e in zip(D["ns"], D["oss"], D["es"]):
            for _orb in _os:
                logger.info(f"atom-{_n} {_orb}")
                dos_plotter.add_dos(
                    f"{_e}(atom-{_n}) {_orb}",  # label
                    dos_data.get_site_orbital_dos(  # type: ignore
                        dos_data.structure[int(_n)],  # type: ignore
                        getattr(Orbital, _orb),
                    ),
                )
        dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D


def s5_5(language: Literal["EN", "CN"]) -> dict:
    "Projection of density of states onto split d orbitals (t2g, eg) on different atoms"
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import DosPlotter as cli_DosPlotter

        from dspawpy.io.read import get_dos_data as cli_get_dos_data

        for D in list_ds:
            if D["menu"] == 52:
                dos_data = cli_get_dos_data(D["inf"])
                if D["shift"] == "y":
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                else:
                    dos_plotter = cli_DosPlotter(stack=False, zero_at_efermi=True)
                atom_indices = [int(ai) for ai in D["ais"].split()]
                for atom_index in atom_indices:
                    dos_plotter.add_dos_dict(
                        dos_data.get_site_t2g_eg_resolved_dos(  # type: ignore
                            dos_data.structure[atom_index]  # type: ignore
                        )
                    )

                dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
                save_figure_dpi300(D["figure"])
        D = {}
    else:
        D, DosPlotter, get_dos_data = pre_ele_pdos(language)
        D["menu"] = 55
        dos_data = get_dos_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        if D["shift"] == "y":
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        else:
            dos_plotter = DosPlotter(stack=False, zero_at_efermi=True)
        logger.debug("plotter initialized")
        logger.info(dos_data.structure)  # type: ignore
        sites = dos_data.structure.sites  # type: ignore
        numbers = [str(e) for e in range(len(sites))]
        D["ais"] = get_input(Dselect[language][6], numbers)

        atom_indices = [int(ai) for ai in D["ais"].split()]
        for atom_index in atom_indices:
            dos_plotter.add_dos_dict(  # type: ignore
                dos_data.get_site_t2g_eg_resolved_dos(dos_data.structure[atom_index]),  # type: ignore
            )

        dos_plotter.get_plot(xlim=D["xlims"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D


def s5_6(language: Literal["EN", "CN"]) -> dict:
    """d-centralized analysis with center"""
    if auto_test_cli:
        from dspawpy.io.read import get_dos_data as cli_get_dos_data
        from dspawpy.io.utils import d_band as cli_d_band

        for D in list_ds:
            if D["menu"] == 56:
                dos_data = cli_get_dos_data(D["inf"])
                os.makedirs(
                    os.path.dirname(os.path.abspath(D["outfile"])), exist_ok=True
                )
                with open(D["outfile"], "w") as f:
                    for spin in dos_data.densities:
                        # up, down = (1, -1)
                        if spin.value == 1:
                            s = "up"
                        elif spin.value == -1:
                            s = "down"
                        else:
                            raise ValueError(f"Unknown spin: {spin}")
                        logger.info("spin=", s)
                        f.write(f"spin={s}\n")
                        c = cli_d_band(spin, dos_data)
                        f.write(str(c) + "\n")
                        logger.info(c)
        D = {}
    else:

        def imp():
            global get_dos_data, d_band

            from dspawpy.io.read import get_dos_data
            from dspawpy.io.utils import d_band

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 56
        D["inf"] = prompt(Dio[language]["pdos"], completer=pc)
        D["outfile"] = prompt(Dio[language]["txt"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        dos_data = get_dos_data(D["inf"])
        os.makedirs(os.path.dirname(os.path.abspath(D["outfile"])), exist_ok=True)
        with open(D["outfile"], "w") as f:
            for spin in dos_data.densities:
                # up, down = (1, -1)
                if spin.value == 1:
                    s = "up"
                elif spin.value == -1:
                    s = "down"
                else:
                    raise ValueError(f"Unknown spin: {spin}")
                logger.info("spin=", s)
                f.write(f"spin={s}\n")
                c = d_band(spin, dos_data)
                f.write(str(c) + "\n")
                logger.info(c)

    return D



from typing import Optional

from loguru import logger


def _zip_folder(folder_path: str, output_path: str):
    import os

    absdir1 = os.path.abspath(folder_path)
    absdir2 = os.path.abspath(output_path)
    import zipfile

    with zipfile.ZipFile(absdir2, "w", zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(absdir1):
            for file in files:
                file_path = os.path.join(root, file)
                zipf.write(file_path, os.path.relpath(file_path, absdir1))


def get_distance(spo1, spo2, lat1, lat2):
    """Calculate the distance between two structures based on their fractional coordinates and cell parameters

    Parameters
    ----------
    spo1 : np.ndarray
        Scores coordinate list 1
    spo2 : np.ndarray
        Fractional coordinate list 2
    lat1 : np.ndarray
        Cell 1
    lat2 : np.ndarray
        Cell 2

    Returns
    -------
    float
        Distance

    Examples
    --------
    First, read the structure information

    >>> from dspawpy.io.structure import read
    >>> s1 = read('tests/2.15/01/structure01.as')[0]
    >>> s2 = read('tests/2.15/02/structure02.as')[0]

    Calculate the distance between two configurations

    >>> from dspawpy.diffusion.nebtools import get_distance
    >>> dist = get_distance(s1.frac_coords, s2.frac_coords, s1.lattice.matrix, s2.lattice.matrix)
    >>> print('The distance between the two configurations is:', dist, 'Angstrom')
    The distance between the two configurations is: 0.476972808803491 Angstrom

    """
    import numpy as np

    diff_spo = spo1 - spo2  # Fractional coordinate difference
    avglatv = 0.5 * (lat1 + lat2)  # Average lattice vector
    pbc_diff_spo = set_pbc(diff_spo)  # Cartesian coordinate difference
    # Dot product of fractional coordinates with average lattice, converted back to Cartesian coordinates
    pbc_diff_pos = np.dot(pbc_diff_spo, avglatv)  # Cartesian coordinate difference
    distance = np.sqrt(np.sum(pbc_diff_pos**2))

    return distance


def get_neb_subfolders(directory: str = ".", return_abs: bool = False):
    r"""List the subfolder names under the directory path and sort them by numerical size

    ```

        Only retain NEB subfolder paths that are of the form 00, 01 numeric types

        Parameters
        ----------
        subfolders:
            List of subfolder names
        return_abs : bool, optional
            Whether to return absolute paths, default is False

        Returns
        -------
        subfolders:
            Sorted list of subfolder names

        Examples
        --------
        >>> from dspawpy.diffusion.nebtools import get_neb_subfolders
        >>> directory = 'tests/2.15'
        >>> get_neb_subfolders(directory)
        ['00', '01', '02', '03', '04']

    """
    import os

    absdir = os.path.abspath(directory)
    raw_subfolders = next(os.walk(absdir))[1]
    subfolders = []
    for subfolder in raw_subfolders:
        try:
            assert 0 <= int(subfolder) < 100
            subfolders.append(subfolder)
        except Exception:
            pass
    subfolders.sort()  # Sort in ascending order
    if return_abs:
        subfolders = [
            os.path.abspath(os.path.join(absdir, subfolder)) for subfolder in subfolders
        ]
    return subfolders


def plot_barrier(
    datafile: str = "neb.h5",
    directory: Optional[str] = None,
    ri: Optional[float] = None,
    rf: Optional[float] = None,
    ei: Optional[float] = None,
    ef: Optional[float] = None,
    method: str = "PchipInterpolator",
    figname: Optional[str] = "neb_barrier.png",
    show: bool = True,
    raw: bool = False,
    verbose: bool = False,
    **kwargs,
):
    r"""Call the scipy.interpolate interpolation algorithm to fit the NEB barrier and plot

    Parameters
    ----------
    datafile:
        Path to neb.h5 or neb.json file
    directory:
        NEB calculation path
    ri:
        Initial reaction coordinate
    rf:
        Final state reaction coordinate
    ei:
        Initial state self-consistent energy
    ef:
        Final state self-consistent energy
    method : str, optional
        Interpolation algorithm, default 'PchipInterpolator'
    figname : str, optional
        Barrier image name, default 'neb_barrier.png'
    show : bool, optional
        Whether to display the interactive interface, default True
    raw : bool, optional
        Whether to return plotting data to CSV

    Raises
    ------
    ImportError
        The specified interpolation algorithm does not exist in scipy.interpolate
    ValueError
        The parameters passed to the interpolation algorithm do not meet the requirements of the algorithm

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import plot_barrier
    >>> import matplotlib.pyplot as plt

    Comparing different interpolation algorithms

    >>> plot_barrier(directory='tests/2.15', method='interp1d', kind=2, figname=None, show=False)
    >>> plot_barrier(directory='tests/2.15', method='interp1d', kind=3, figname=None, show=False)
    >>> plot_barrier(directory='tests/2.15', method='CubicSpline', figname=None, show=False)
    >>> plot_barrier(directory='tests/2.15', method='pchip', figname='tests/outputs/doctest/barrier_comparison.png', show=False) # doctest: +ELLIPSIS
    ==> .../barrier_comparison...png

    Attempt to read neb.h5 file or neb.json file

    >>> plot_barrier(datafile='tests/2.15/neb.h5', method='pchip', figname='tests/outputs/doctest/barrier_h5.png', show=False) # doctest: +ELLIPSIS
    ==> .../barrier_h5...png
    >>> plot_barrier(datafile='tests/2.15/neb.json', method='pchip', figname='tests/outputs/doctest/barrier_json.png', show=False) # doctest: +ELLIPSIS
    ==> .../barrier_json...png

    """
    import os

    import numpy as np

    if directory is not None:
        # read data
        subfolders, resort_mfs, rcs, ens, dEs = _getef(os.path.abspath(directory))

    elif datafile:
        from dspawpy.io.utils import get_absfile

        absfile = get_absfile(
            datafile,
            task="neb",
            verbose=verbose,
        )  # -> return either .h5 or .json
        if absfile.endswith(".h5"):
            from dspawpy.io.read import load_h5

            neb = load_h5(absfile)
            if "/BarrierInfo/ReactionCoordinate" in neb.keys():
                reaction_coordinate = neb["/BarrierInfo/ReactionCoordinate"]
                energy = neb["/BarrierInfo/TotalEnergy"]
            else:  # old version
                reaction_coordinate = neb["/Distance/ReactionCoordinate"]
                energy = neb["/Energy/TotalEnergy"]
        elif absfile.endswith(".json"):
            with open(absfile) as fin:
                from json import load

                neb = load(fin)
            if "BarrierInfo" in neb.keys():
                reaction_coordinate = neb["BarrierInfo"]["ReactionCoordinate"]
                energy = neb["BarrierInfo"]["TotalEnergy"]
            else:  # old version
                reaction_coordinate = neb["Distance"]["ReactionCoordinate"]
                energy = neb["Energy"]["TotalEnergy"]
        else:
            raise ValueError("only h5 and json file are supported")

        x = reaction_coordinate  # No need to sum when reading from neb.h5/json

        y = [x - energy[0] for x in energy]
        # initial and final info
        if ri is not None:  # add initial reaction coordinate
            x.insert(0, ri)
        if rf is not None:  # add final reaction coordinate
            x.append(rf)

        if ei is not None:  # add initial energy
            y.insert(0, ei)
        if ef is not None:  # add final energy
            y.append(ef)

        rcs = np.asarray(x)
        dEs = np.asarray(y)

    else:
        raise ValueError("Please specify directory or datafile!")

    # import scipy interpolater
    try:
        interpolate_method = getattr(
            __import__("scipy.interpolate", fromlist=[method]),
            method,
        )
    except Exception:
        raise ImportError(f"No scipy.interpolate.{method} method！")
    # call the interpolater to interpolate with given kwargs
    try:
        inter_f = interpolate_method(rcs, dEs, **kwargs)
    except Exception:
        raise ValueError(f"Please check whether {kwargs} is valid for {method}！")

    xnew = np.linspace(rcs[0], rcs[-1], 100)
    ynew = inter_f(xnew)

    if raw:
        import polars as pl

        pl.DataFrame({"x_raw": rcs, "y_raw": dEs}).write_csv(
            "raw_xy.csv",
        )
        pl.DataFrame({"x_interpolated": xnew, "y_interpolated": ynew}).write_csv(
            "raw_interpolated_xy.csv",
        )

    # plot
    import matplotlib.pyplot as plt

    if kwargs:
        plt.plot(xnew, ynew, label=method + str(kwargs))
    else:
        plt.plot(xnew, ynew, label=method)
    plt.scatter(rcs, dEs, c="r")
    plt.xlabel("Reaction Coordinate (Å)")
    plt.ylabel("Energy (eV)")
    plt.legend()

    plt.tight_layout()
    # save and show
    if figname:
        absfig = os.path.abspath(figname)
        os.makedirs(os.path.dirname(absfig), exist_ok=True)
        plt.savefig(absfig, dpi=300)
        print(f"==> {absfig}")
    if show:
        plt.show()


def plot_neb_converge(
    neb_dir: str,
    image_key: str = "01",
    show: bool = True,
    figname: str = "neb_conv.png",
    raw=False,
    verbose: bool = False,
):
    """Specify the NEB calculation path, plot the NEB convergence process graph

    Parameters
    ----------
    neb_dir:
        Path to the neb.h5 / neb.json file or the folder containing the neb.h5 / neb.json file
    image_key:
        The index of the configuration, default "01"
    show:
        Interactive plotting
    image_name:
        NEB convergence plot name, default "neb_conv.png"
    raw:
        Whether to output plot data to a CSV file

    Returns
    -------
    ax1, ax2 : matplotlib.axes.Axes
        Two Axes objects for the subplots

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import plot_neb_converge
    >>> result = plot_neb_converge(neb_dir='tests/2.15', image_key='01', figname='tests/outputs/doctest/neb_converge1.png',show=False) # doctest: +ELLIPSIS
    ==> .../neb_converge1...png
    >>> result = plot_neb_converge(neb_dir='tests/2.15/neb.h5', image_key='02', figname='tests/outputs/doctest/neb_converge2.png',show=False) # doctest: +ELLIPSIS
    ==> .../neb_converge2...png
    >>> result = plot_neb_converge(neb_dir='tests/2.15/neb.json', image_key='03', figname='tests/outputs/doctest/neb_converge3.png',show=False, raw=True) # doctest: +ELLIPSIS
    ==> .../neb_converge3...png

    """
    import os

    import numpy as np

    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(neb_dir, "neb", verbose=verbose)
    if absfile.endswith("h5"):
        import h5py

        neb_total = h5py.File(absfile)
        # new output (>=2022B)
        if "/LoopInfo/01/MaxForce" in neb_total.keys():
            maxforce = np.asarray(neb_total.get("/LoopInfo/" + image_key + "/MaxForce"))
        else:  # old output
            maxforce = np.asarray(
                neb_total.get("/Iteration/" + image_key + "/MaxForce")
            )

        if "/LoopInfo/01/TotalEnergy" in neb_total.keys():  # new output (>=2022B)
            total_energy = np.asarray(
                neb_total.get("/LoopInfo/" + image_key + "/TotalEnergy"),
            )
        else:  # old output
            total_energy = np.asarray(
                neb_total.get("/Iteration/" + image_key + "/TotalEnergy"),
            )

    elif absfile.endswith("json"):
        with open(absfile) as fin:
            from json import load

            neb_total = load(fin)
        if "LoopInfo" in neb_total.keys():
            neb = neb_total["LoopInfo"][image_key]
        else:
            neb = neb_total["Iteration"][image_key]
        maxforce = []
        total_energy = []
        for n in neb:
            maxforce.append(n["MaxForce"])
            total_energy.append(n["TotalEnergy"])

        maxforce = np.asarray(maxforce)
        total_energy = np.asarray(total_energy)

    else:
        raise ValueError("Only h5 and json file are supported")

    x = np.arange(len(maxforce))

    force = maxforce
    energy = total_energy

    if raw:
        import polars as pl

        pl.DataFrame({"x": x, "force": force, "energy": energy}).write_csv(
            "neb_conv.csv",
        )

    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(x, force, label="Max Force", c="black")
    ax1.set_xlabel("Number of ionic step")
    ax1.set_ylabel("Force (eV/Å)")
    ax2 = ax1.twinx()
    ax2.plot(x, energy, label="Energy", c="r")  # type: ignore
    ax2.set_xlabel("Number of ionic step")
    ax2.set_ylabel("Energy (eV)")
    ax2.ticklabel_format(
        useOffset=False
    )  # Display absolute values on the y-axis instead of relative values
    fig.legend(loc=1, bbox_to_anchor=(1, 1), bbox_transform=ax1.transAxes)

    plt.tight_layout()
    # save and show
    if figname:
        absfig = os.path.abspath(figname)
        os.makedirs(os.path.dirname(absfig), exist_ok=True)
        plt.savefig(absfig, dpi=300)
        print(f"==> {absfig}")
    if show:
        plt.show()

    return ax1, ax2


def printef(directory: str):
    """Print the energies and forces of each configuration during NEB calculations

    Parameters
    ----------
    directory:
        Directory for NEB calculations, defaulting to the current directory

    Returns
    -------
    Prints the energy and forces for each configuration

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import printef
    >>> printef(directory='tests/2.15')
    shape: (5, 5)
    ┌────────────┬─────────────┬──────────┬───────────────┬──────────┐
    │ FolderName ┆ Force(eV/Å) ┆ RC(Å)    ┆ Energy(eV)    ┆ E-E0(eV) │
    ╞════════════╪═════════════╪══════════╪═══════════════╪══════════╡
    │ 00         ┆ 0.180272    ┆ 0.0      ┆ -39637.097656 ┆ 0.0      │
    │ 01         ┆ 0.014094    ┆ 0.542789 ┆ -39637.019531 ┆ 0.079814 │
    │ 02         ┆ 0.026337    ┆ 1.0868   ┆ -39636.878906 ┆ 0.218265 │
    │ 03         ┆ 0.024798    ┆ 1.588367 ┆ -39637.0      ┆ 0.100043 │
    │ 04         ┆ 0.234429    ┆ 2.089212 ┆ -39637.089844 ┆ 0.008414 │
    └────────────┴─────────────┴──────────┴───────────────┴──────────┘
    """
    import numpy as np

    from dspawpy.diffusion.nebtools import _getef

    arr = np.asarray(_getef(directory))
    data = {
        "FolderName": arr[0],
        "Force(eV/Å)": arr[1],
        "RC(Å)": arr[2],
        "Energy(eV)": arr[3],
        "E-E0(eV)": arr[4],
    }
    import polars as pl

    _df = pl.DataFrame(data)
    col_names = [col_name for col_name in _df.columns if col_name != "FolderName"]
    df = _df.with_columns(pl.col(col_names).cast(pl.Float32, strict=False))
    print(df)


def restart(directory: str = ".", output: str = "bakfile"):
    """Archive and compress old NEB tasks, and prepare for continuation at the original path

    Parameters
    ----------
    directory:
        Old NEB task path, default current path
    output:
        Backup folder path, default is to create a bakfile folder in the current path for backup;
        Alternatively, you can specify any path, but it cannot be the same as the current path

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import restart
    >>> from shutil import copytree
    >>> copytree('tests/2.15', 'tests/outputs/doctest/neb4bk2', dirs_exist_ok=True)
    'tests/outputs/doctest/neb4bk2'
    >>> restart(directory='tests/outputs/doctest/neb4bk2', output='tests/outputs/doctest/neb_backup') # doctest: +ELLIPSIS
    ==> .../neb_backup...

    The preparation for the continuation calculation may take a long time to complete, please be patient

    """
    import os
    import shutil

    from dspawpy.io.write import handle_duplicated_output

    absdir = os.path.abspath(directory)
    absolute_output = handle_duplicated_output(os.path.abspath(output))

    subfolders = get_neb_subfolders(absdir, return_abs=True)  # Get subfolder paths
    os.makedirs(absolute_output, exist_ok=True)  # Create the bakfile folder
    # First process subfolders 00, 01...
    for subfolder_old in subfolders:
        folder_index = subfolder_old.split("/")[-1]  # 00，01...
        subfolder_back = os.path.join(
            absolute_output, folder_index
        )  # Backup subfolders to this location
        shutil.move(subfolder_old, subfolder_back)
        os.makedirs(subfolder_old, exist_ok=True)  # Original folder cleared

        latestStructureFile = f"{subfolder_back}/latestStructure{folder_index}.as"
        structureFile = f"{subfolder_back}/structure{folder_index}.as"

        # Copy the structure file to the original path for continuation calculation; use the latest structure file (latestStructureFile) if it exists, otherwise use s_in_old (s), and raise an error if neither exists
        s_in_old = f"{subfolder_old}/structure{folder_index}.as"
        if os.path.isfile(latestStructureFile):
            shutil.copy(latestStructureFile, s_in_old)
        elif os.path.isfile(structureFile):
            shutil.copy(structureFile, s_in_old)
        else:
            raise FileNotFoundError(
                f"{latestStructureFile} and {structureFile} do not exist!"
            )

        # Temporarily placed in the backup main path; if neither exists, an error has already been raised earlier
        ls_bk = os.path.join(absdir, f"latestStructure{folder_index}.as")
        s_bk = os.path.join(absdir, f"structure{folder_index}.as")
        if os.path.isfile(latestStructureFile):
            shutil.copy(latestStructureFile, ls_bk)
        if os.path.isfile(structureFile):
            shutil.copy(structureFile, s_bk)

        # Handle subfolders in the backup path
        zf = f"{absolute_output}/{folder_index}.zip"
        _zip_folder(subfolder_back, zf)  # Compress subfolder
        # Empty the backup subfolder
        for f in os.listdir(subfolder_back):
            fpath = os.path.join(subfolder_back, f)
            if os.path.isfile(fpath):
                os.remove(fpath)
            else:
                shutil.rmtree(fpath)

        # Move the archive and structure files in
        shutil.move(zf, f"{subfolder_back}/{folder_index}.zip")
        if os.path.isfile(ls_bk) and os.path.isfile(s_bk):
            shutil.move(ls_bk, f"{subfolder_back}/latestStructure{folder_index}.as")
            shutil.move(s_bk, f"{subfolder_back}/structure{folder_index}.as")
        elif os.path.isfile(ls_bk):
            shutil.move(ls_bk, f"{subfolder_back}/latestStructure{folder_index}.as")
        elif os.path.isfile(s_bk):
            shutil.move(s_bk, f"{subfolder_back}/structure{folder_index}.as")
        else:
            raise FileNotFoundError(f"Both {ls_bk} and {s_bk} not found")

    # Process individual files in the main directory of the old NEB folder
    # Backup neb.h5, neb.json, and DS-PAW.log
    tmp_dir = os.path.join(absolute_output, "tmp")
    os.makedirs(tmp_dir, exist_ok=True)
    if os.path.isfile(f"{absdir}/neb.h5"):
        shutil.move(f"{absdir}/neb.h5", f"{tmp_dir}/neb.h5")

    if os.path.isfile(f"{absdir}/neb.json"):
        shutil.move(f"{absdir}/neb.json", f"{tmp_dir}/neb.json")

    if len(os.listdir(tmp_dir)) > 0:  # If there are data files
        _zip_folder(tmp_dir, f"{absolute_output}/neb_data.zip")
        for f in os.listdir(tmp_dir):
            os.remove(os.path.join(tmp_dir, f))
        os.removedirs(tmp_dir)

    if os.path.isfile(f"{absdir}/DS-PAW.log"):
        shutil.move(f"{absdir}/DS-PAW.log", f"{absolute_output}/DS-PAW.log")

    print(f"==> {absolute_output}")


def set_pbc(spo):
    """Shift fractional coordinate components into the [-0.5, 0.5) interval according to periodic boundary conditions.

    Parameters
    ----------
    spo : np.ndarray or list
        A list of fractional coordinates

    Returns
    -------
    pbc_spo : np.ndarray
        A list of fractional coordinates satisfying periodic boundary conditions

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import set_pbc
    >>> set_pbc([-0.6, 1.2, 2.3])
    array([0.4, 0.2, 0.3])

    """
    # wrap into [-0.5, 0.5)
    import numpy as np

    pbc_spo = np.mod(np.asarray(spo) + 0.5, 1.0) - 0.5

    return pbc_spo


def summary(
    directory: str = ".",
    raw=False,
    show_converge=False,
    outdir: Optional[str] = None,
    **kwargs,
):
    r"""Summary of NEB task completion, execute the following steps in order:

    - 1. Print the forces, reaction coordinates, energy, and energy differences from the initial configuration for each structure
    - 2. Plot the energy barrier diagram
    - 3. Plot and save the convergence processes of energy and forces during the structure optimization

    Parameters
    ----------
    directory:
        NEB path, default to the current path
    raw:
        Whether to save the plot data to a CSV file
    show_converge:
        Whether to display energy and force convergence plots of the structural optimization process, default is not displayed
    outdir:
        Path to save the convergence process figure, default to directory
    **kwargs : dict
        Parameters passed to plot_barrier

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import summary
    >>> directory = 'tests/2.15' # Path for NEB calculation, default to current path
    >>> summary(directory, show=False, figname='tests/outputs/doctest/neb_barrier.png') # doctest: +ELLIPSIS
    shape: (5, 5)
    ┌────────────┬─────────────┬──────────┬───────────────┬──────────┐
    │ FolderName ┆ Force(eV/Å) ┆ RC(Å)    ┆ Energy(eV)    ┆ E-E0(eV) │
    ╞════════════╪═════════════╪══════════╪═══════════════╪══════════╡
    │ 00         ┆ 0.180272    ┆ 0.0      ┆ -39637.097656 ┆ 0.0      │
    │ 01         ┆ 0.014094    ┆ 0.542789 ┆ -39637.019531 ┆ 0.079814 │
    │ 02         ┆ 0.026337    ┆ 1.0868   ┆ -39636.878906 ┆ 0.218265 │
    │ 03         ┆ 0.024798    ┆ 1.588367 ┆ -39637.0      ┆ 0.100043 │
    │ 04         ┆ 0.234429    ┆ 2.089212 ┆ -39637.089844 ┆ 0.008414 │
    └────────────┴─────────────┴──────────┴───────────────┴──────────┘
    ==> .../neb_barrier...png
    ==> .../converge...png
    ==> .../converge...png
    ==> .../converge...png

    >>> summary(directory, show=False, figname='tests/outputs/doctest/neb_barrier.png', outdir="tests/outputs/doctest/neb_summary") # doctest: +ELLIPSIS
    shape: (5, 5)
    ┌────────────┬─────────────┬──────────┬───────────────┬──────────┐
    │ FolderName ┆ Force(eV/Å) ┆ RC(Å)    ┆ Energy(eV)    ┆ E-E0(eV) │
    ╞════════════╪═════════════╪══════════╪═══════════════╪══════════╡
    │ 00         ┆ 0.180272    ┆ 0.0      ┆ -39637.097656 ┆ 0.0      │
    │ 01         ┆ 0.014094    ┆ 0.542789 ┆ -39637.019531 ┆ 0.079814 │
    │ 02         ┆ 0.026337    ┆ 1.0868   ┆ -39636.878906 ┆ 0.218265 │
    │ 03         ┆ 0.024798    ┆ 1.588367 ┆ -39637.0      ┆ 0.100043 │
    │ 04         ┆ 0.234429    ┆ 2.089212 ┆ -39637.089844 ┆ 0.008414 │
    └────────────┴─────────────┴──────────┴───────────────┴──────────┘
    ==> .../neb_barrier...png
    ==> .../converge...png
    ==> .../converge...png
    ==> .../converge...png

    If inifin=False, the user must place a converged scf.h5 or system.json in the initial and final state subfolders.

    """
    import os

    # 1. Plot the energy barrier curve
    absdir = os.path.abspath(directory)
    printef(absdir)

    # 2. Print the forces, reaction coordinates, energy, and energy difference from the initial configuration for each structure
    import matplotlib.pyplot as plt

    plt.clf()  # Clear the canvas before plotting
    plot_barrier(directory=absdir, raw=raw, **kwargs)

    # 3. Plot and save the convergence processes of energy and forces during the structure optimization to each configuration folder
    subfolders = get_neb_subfolders(absdir)
    for subfolder in subfolders[1 : len(subfolders) - 1]:
        if outdir:
            absolute_output = os.path.abspath(outdir)
            os.makedirs(os.path.join(absolute_output, subfolder), exist_ok=True)
            pngfile = f"{absolute_output}/{subfolder}/converge.png"
        else:
            pngfile = (
                f"tests/outputs/doctest/{subfolder}/converge.png"
            )

        plot_neb_converge(
            neb_dir=absdir,
            image_key=subfolder,
            figname=pngfile,
            raw=raw,
            show=show_converge,
        )
    plt.clf()


def write_movie_json(
    preview: bool = False,
    directory: str = ".",
    step: int = -1,
    dst: Optional[str] = None,
):
    DeprecationWarning("Please use write_json_chain() instead")
    write_json_chain(preview=preview, directory=directory, step=step, dst=dst)


def write_json_chain(
    preview: bool = False,
    directory: str = ".",
    step: int = -1,
    dst: Optional[str] = None,
    ignorels=False,
):
    r"""Read information after NEB calculation or initial interpolation, and save it as neb_chain*.json files

    View the structure and other information by opening the file with Device Studio.

    Parameters
    ----------
    preview:
        Whether to enable preview mode, default is no
    directory:
        Directory where the computation results are located. Default to the current path
    step:
        - Ion step number. Default is -1, which reads the entire NEB calculation process information; at this time, the priority is latestStructure.as > h5 > json
        - 0 indicates the initial insertion structure (uncompleted ionic step);
        - 1 represents the first ionic step, and so on
    dst:
        Save path, default to directory
    ignorels:
        Whether to ignore the latestStructure.as file when step=-1, default is no

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import write_json_chain

    To observe the entire trajectory change process after NEB computation, simply specify the NEB computation path.

    >>> write_json_chain(directory='tests/2.15', dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    structure info collected from latestStructure.as
    ==> .../neb_chain_last...json

    To observe the structure of the nth ion step after NEB computation, set step to n, noting that step is counted starting from 1.

    >>> write_json_chain(directory='tests/2.15', step=1, dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_1...json

    If the specified step number exceeds the actual ion steps completed by NEB, it will be automatically adjusted to the final step.

    >>> write_json_chain(directory='tests/2.15', step=10, dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_10...json
    >>> write_json_chain(directory='tests/2.15', step=100, dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_last...json

    Additionally, to preview the initial insertion structure, set preview to True and specify the directory as the main path of the NEB calculation

    >>> write_json_chain(preview=True, directory='tests/2.15', dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_init...json

    """
    import os

    absdir = os.path.abspath(directory)
    if preview or step == 0:  # preview mode
        raw = get_raw_from_structure(absdir)
    elif step == -1:  # priority latestStructure.as > h5 > json
        if ignorels:
            raw = _from_h5(absdir, step)
        else:
            try:  # read ls.as
                raw = _from_structures(absdir, ls=True)
                print("structure info collected from latestStructure.as")
            except FileNotFoundError:
                raw = _from_h5(absdir, step)
    else:
        raw = _from_h5(absdir, step)

    new = []
    assert raw is not None
    if dst is not None:
        abs_dst = os.path.abspath(dst)
        os.makedirs(abs_dst, exist_ok=True)
        new.append(f"{abs_dst}/{raw[0]}")
        for i in range(1, len(raw)):
            new.append(raw[i])
        _dump_neb_chain_json(new)
    else:
        _dump_neb_chain_json(raw)


def write_xyz(
    preview: bool = False,
    directory: str = ".",
    step: int = -1,
    dst: Optional[str] = None,
):
    DeprecationWarning("Please use write_xyz_chain() instead")
    write_xyz_chain(preview=preview, directory=directory, step=step, dst=dst)


def get_raw_from_structure(dire: str):
    try:
        raw = _from_structures(dire)
        return raw
    except FileNotFoundError:
        print("No structure file")
    except Exception as e:
        print(e)


def write_xyz_chain(
    preview: bool = False,
    directory: str = ".",
    step: int = -1,
    dst: Optional[str] = None,
    ignorels=False,
):
    r"""Write the NEB structure chains to an xyz trajectory file for visualization

    Parameters
    ----------
    preview:
        Whether to enable preview mode, default is no
    directory:
        Directory where the calculation results are located. Default to the current path
    step:
        - Ion step number. Default is -1, which reads the entire NEB calculation process information; at this time, the priority is latestStructure.as > h5 > json
        - 0 indicates the initial inserted structure (not completed ion step);
        - 1 represents the first ionic step, and so on
    dst:
        Save path, default to directory
    ignorels:
        Whether to ignore the latestStructure.as file when step=-1, default is no

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import write_xyz_chain

    To observe the entire process of trajectory changes after NEB calculation, simply specify the NEB calculation path.

    >>> write_xyz_chain(directory='tests/2.15', dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    structure info collected from latestStructure.as
    ==> .../neb_chain_last...xyz

    To observe the structure of the nth ion step after NEB calculation, set step to n, note that step is counted starting from 1

    >>> write_xyz_chain(directory='tests/2.15', step=1, dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_1...xyz

    If the specified step number exceeds the actual ion steps completed by NEB, it will be automatically adjusted to the final step.

    >>> write_xyz_chain(directory='tests/2.15', step=10, dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_10...xyz

    Additionally, to preview the initial insertion structure, set preview to True and specify the directory as the main path of the NEB calculation

    >>> write_xyz_chain(preview=True, directory='tests/2.15', dst='tests/outputs/doctest') # doctest: +ELLIPSIS
    ==> .../neb_chain_init...xyz

    """
    import os

    absdir = os.path.abspath(directory)
    if preview or step == 0:  # preview mode, write neb_chain_init.xyz from structure.as
        raw = get_raw_from_structure(absdir)
    elif step == -1:  # Priority: latestStructure.as > h5 > json
        if ignorels:
            raw = _from_h5(absdir, step)
        else:
            try:  # read ls.as
                raw = _from_structures(absdir, ls=True)
                print("structure info collected from latestStructure.as")
            except FileNotFoundError:
                raw = _from_h5(absdir, step)
    else:
        raw = _from_h5(absdir, step)

    new = []
    assert raw is not None
    if dst is not None:
        abs_dst = os.path.abspath(dst)
        os.makedirs(abs_dst, exist_ok=True)
        new.append(f"{abs_dst}/{raw[0]}")
        for i in range(1, len(raw)):
            new.append(raw[i])
        _dump_neb_xyz(new)
    else:
        _dump_neb_xyz(raw)


def _dump_neb_xyz(raw):
    """Dump the JSON file to the output based on the previously collected data lists"""
    import os

    import numpy as np

    (
        output,
        subfolders,
        step,
        MaxForces,
        TotalEnergies,
        Poses,  # Nimage x Natom x 3 , read
        Latvs,  # Nimage x 9
        Elems,  # Nimage x Natom
        Fixs,  # Natom x 3
        reactionCoordinates,
        totalEnergies,
        maxForces,
        tangents,
        iDirects,
    ) = raw

    # Write to file
    xyzfile = output[:-5] + ".xyz"
    absolute_output = os.path.abspath(xyzfile)
    Nstep = len(subfolders)  # Select ion step, display configuration chain
    with open(absolute_output, "w") as f:
        # Nstep
        for n in range(Nstep):
            elements = Elems[n]  # For each configuration
            # Number of atoms remains constant, this is the total number of elements without merging
            f.write("%d\n" % len(elements))
            # lattice
            f.write(
                'Lattice="%f %f %f %f %f %f %f %f %f" Properties=species:S:1:pos:R:3 pbc="T T T"\n'
                % (
                    Latvs[n, 0],
                    Latvs[n, 1],
                    Latvs[n, 2],
                    Latvs[n, 3],
                    Latvs[n, 4],
                    Latvs[n, 5],
                    Latvs[n, 6],
                    Latvs[n, 7],
                    Latvs[n, 8],
                ),
            )
            lat = Latvs[n].reshape(3, 3)
            if iDirects[n]:
                Poses[n] = np.dot(Poses[n], lat)

            # position and element
            for i in range(len(elements)):
                f.write(
                    "%s %f %f %f\n"
                    % (elements[i], Poses[n, i, 0], Poses[n, i, 1], Poses[n, i, 2]),
                )

    print(f"==> {absolute_output}")


def _from_structures(directory: str, ls: bool = False):
    """Read structure information from structure00.as, structure01.as, ..., etc.,
    Write to neb_chain_init so that it can be observed with DeviceStudio

    Parameters
    ----------
    directory:
        NEB calculation path, default to the current path
    ls:
        Whether it is the latest configuration

    Returns
    -------
    Arrays for JSON files

    """
    import os

    absdir = os.path.abspath(directory)
    if ls:
        output = "neb_chain_last.json"
    else:
        output = "neb_chain_init.json"
    step = 1

    subfolders = get_neb_subfolders(absdir)
    nimage = len(subfolders)
    import numpy as np

    reactionCoordinates = np.zeros(shape=nimage)  # optional
    totalEnergies = np.zeros(shape=nimage)  # optional
    maxForces = np.zeros(shape=nimage)  # optional
    tangents = np.zeros(shape=nimage)  # optional
    MaxForces = np.zeros(shape=(nimage, step + 1))  # optional
    TotalEnergies = np.zeros(shape=(nimage, step + 1))  # optional

    Poses = []  # nimage x Natom x 3 , read
    Elems = []  # nimage x Natom, read
    Latvs = []  # nimage x 9, read

    iDirects = []  # read coordinate type
    from dspawpy.io.structure import read

    for i, folder in enumerate(subfolders):
        if ls and i != 0 and i != len(subfolders) - 1:
            structure_path = os.path.join(absdir, folder, f"latestStructure{folder}.as")
        else:
            structure_path = os.path.join(absdir, folder, f"structure{folder}.as")
        if not os.path.isfile(structure_path):
            raise FileNotFoundError(f"No {structure_path}！")
        structure = read(structure_path, task="free")[0]
        ele = [str(i) for i in structure.species]
        lat = structure.lattice.matrix
        Elems.append(ele)
        Latvs.append(lat)
        with open(structure_path) as f:
            lines = f.readlines()
            coordinateType = lines[6].split()[0]
            if coordinateType == "Direct":
                iDirect = True
                Poses.append(structure.frac_coords)
            elif coordinateType == "Cartesian":
                iDirect = False
                Poses.append(structure.cart_coords)
            else:
                raise ValueError(
                    f"coordinateType in {structure_path} is neither Direct nor Cartesian!",
                )
            iDirects.append(iDirect)
    Natom = len(Elems[0])

    # reshape data
    Poses = np.asarray(Poses).reshape((nimage, Natom, 3))
    Elems = np.asarray(Elems).reshape((nimage, Natom))
    Latvs = np.asarray(Latvs).reshape((nimage, 9))
    Fixs = np.zeros(shape=(Natom, 3))  # optional

    return (
        output,
        subfolders,
        step,
        MaxForces,
        TotalEnergies,
        Poses,
        Latvs,
        Elems,
        Fixs,
        reactionCoordinates,
        totalEnergies,
        maxForces,
        tangents,
        iDirects,
    )


def _from_h5(directory: str, step: int):
    """Read structure and energy information from h5 files under the NEB path, starting from the first step to the specified step number,
    Write to a JSON file for observation with DeviceStudio.

    Supports hot reading structure information (ignoring other information)

    Parameters
    ----------
    directory:
        NEB path, default to current path
    step:
        number of steps, default -1, read the last configuration

    Returns
    -------
    Arrays for JSON files

    """
    import os

    absdir = os.path.abspath(directory)
    # Initial setup
    neb_h5 = os.path.abspath(os.path.join(absdir, "01", "neb01.h5"))
    from dspawpy.io.read import get_ele_from_h5

    ele = get_ele_from_h5(hpath=neb_h5)
    Natom = len(ele)
    import h5py

    data = h5py.File(neb_h5)
    import numpy as np

    try:
        total_steps = np.asarray(data.get("/NebSize"))[0]
    except Exception:
        print("Reading latest info for unfinished NEB task...")
        try:
            total_steps = np.asarray(data.get("/Structures/FinalStep"))[0]
        except Exception:
            raise ValueError(
                f"No finished ionic step detected, please check {neb_h5} file or wait for NEB task to finished at least one ionic step.",
            )

    if step == -1:
        output = "neb_chain_last.json"
        step = total_steps
    elif step > total_steps:
        output = "neb_chain_last.json"
        logger.warning(
            "specified %s > %s, reading last step info..." % (step, total_steps),
        )
        step = total_steps
    else:
        output = f"neb_chain_{step}.json"

    # ^ Prepare the array framework required for the json file before reading
    subfolders = get_neb_subfolders(absdir)
    nimage = len(subfolders)
    reactionCoordinates = np.zeros(shape=nimage)  # optional
    totalEnergies = np.zeros(
        shape=nimage
    )  # optional, final energies for each configuration
    maxForces = np.zeros(shape=nimage - 2)  # optional
    tangents = np.zeros(shape=nimage - 2)  # optional
    MaxForces = np.zeros(shape=(nimage - 2, step))  # optional
    TotalEnergies = np.zeros(
        shape=(nimage - 2, step),
    )  # optional, energy at each ion step for intermediate configurations
    # Sposes = []  # nimage x Natom x 3 , read
    Sposes = np.empty(shape=(nimage, Natom, 3))  # nimage x Natom x 3 , read
    Elems = []  # nimage x Natom, read
    Latvs = []  # nimage x 9, read
    Fixs = []  # Natom x 3, set

    from dspawpy.io.structure import read

    for i, folder in enumerate(subfolders):
        if folder == subfolders[0] or folder == subfolders[-1]:
            h5_path = os.path.join(absdir, folder, "scf.h5")
            spath = os.path.join(absdir, folder, f"structure{folder}.as")
            if os.path.isfile(h5_path):
                data = h5py.File(h5_path)
                # Does not affect visualization, set directly to 0
                if folder == subfolders[0]:
                    reactionCoordinates[i] = 0
                pos = np.asarray(data.get("/Structures/Step-1/Position")).reshape(
                    -1,
                    3,
                )  # scaled
                lat = np.asarray(data.get("/Structures/Step-1/Lattice"))
                ele = get_ele_from_h5(hpath=h5_path)
                totalEnergies[i] = np.asarray(data.get("/Energy/TotalEnergy0")[0])
            else:
                structure = read(spath, task="neb")[0]
                pos = structure.frac_coords
                ele = [str(i) for i in structure.species]
                lat = structure.lattice.matrix
        else:
            h5_path = os.path.join(absdir, folder, f"neb{folder}.h5")
            data = h5py.File(h5_path)
            # reading...
            try:
                reactionCoordinates[i - 1] = np.asarray(data.get("/Distance/Previous"))[
                    -1
                ]
                maxForces[i - 1] = np.asarray(data.get("/MaxForce"))[-1]
                tangents[i - 1] = np.asarray(data.get("/Tangent"))[-1]
                if folder == subfolders[-2]:
                    reactionCoordinates[i + 1] = np.asarray(data.get("/Distance/Next"))[
                        -1
                    ]
                # read MaxForces and TotalEnergies
                nionStep = np.asarray(data.get("/MaxForce")).shape[0]
                assert step <= nionStep, (
                    f"The number of finished ionic steps is {nionStep}"
                )
                for j in range(step):
                    MaxForces[i - 1, j] = np.asarray(data.get("/MaxForce"))[j]
                    TotalEnergies[i - 1, j] = np.asarray(data.get("/TotalEnergy"))[j]
                totalEnergies[i] = np.asarray(data.get("/Energy/TotalEnergy0")[0])
            except Exception:
                pass  # NEB calculation is not yet complete, but it does not affect reading structure information for visualization
            # read the latest structure for visualization
            pos = np.asarray(data.get(f"/Structures/Step-{step}/Position")).reshape(
                Natom,
                3,
            )  # scaled
            lat = np.asarray(data.get(f"/Structures/Step-{step}/Lattice"))
            ele = get_ele_from_h5(hpath=h5_path)

        Elems.append(ele)
        Sposes[i, :, :] = pos
        Latvs.append(lat)

    if os.path.isfile(os.path.join(absdir, "neb.h5")):
        tdata = h5py.File(os.path.join(absdir, "neb.h5"))
        # atom fix, not lattice
        # ignore this trivial message because it is not necessary for the visualization
        if "/UnrelaxStructure/Image00/Fix" in tdata:
            fix_array = np.asarray(tdata.get("/UnrelaxStructure/Image00/Fix"))
            for fix in fix_array:
                if fix == 0.0:
                    F = False
                elif fix == 1.0:
                    F = True
                else:
                    raise ValueError("Fix must be 0/1")
                Fixs.append(F)
        else:
            Fixs = np.full(shape=(Natom, 3), fill_value=False)
    else:
        Fixs = np.full(shape=(Natom, 3), fill_value=False)

    Elems = np.asarray(Elems).reshape((nimage, Natom))
    Latvs = np.asarray(Latvs).reshape((nimage, 9))
    Fixs = np.asarray(Fixs).reshape((Natom, 3))
    iDirects = [True for i in range(Natom)]  # only output direct coordinates

    return (
        output,
        subfolders,
        step,
        MaxForces,
        TotalEnergies,  #
        Sposes,
        Latvs,
        Elems,
        Fixs,
        reactionCoordinates,
        totalEnergies,
        maxForces,
        tangents,
        iDirects,
    )


def _dump_neb_chain_json(raw):
    """Dump the JSON file to output based on the previously collected data lists"""
    (
        output,
        subfolders,
        step,
        MaxForces,
        TotalEnergies,
        Poses,
        Latvs,
        Elems,
        Fixs,
        reactionCoordinates,
        totalEnergies,
        maxForces,
        tangents,
        iDirects,
    ) = raw

    IterDict = {}
    for s, sf in enumerate(subfolders):
        if sf == subfolders[0] or sf == subfolders[-1]:
            continue
        else:
            Eflist = []
            for _i in range(step):
                ef = {
                    "MaxForce": MaxForces[s - 1, _i],
                    "TotalEnergy": TotalEnergies[s - 1, _i],
                }
                Eflist.append(ef)
                iterDict = {sf: Eflist}  # construct sub-dict
                IterDict.update(iterDict)  # append sub-dict

    RSList = []
    """
    Traverse configurations and atoms (sub-dictionaries) from outside to inside.
    The key-value pair for an atom is: 'Atoms': atom information list
    The atom information list is a list of dictionaries, where each dictionary corresponds to the information of an atom.
    """
    for s, sf in enumerate(subfolders):
        pos = Poses[s]
        lat = Latvs[s]
        elem = Elems[s]
        atoms = []
        for i in range(len(elem)):
            atom = {
                "Element": elem[i],
                "Fix": Fixs[i].tolist(),
                "Mag": [],  # empty
                "Position": pos[i].tolist(),
                "Pot": "",
            }  # empty
            atoms.append(atom)
        if iDirects[s]:
            rs = {"Atoms": atoms, "CoordinateType": "Direct", "Lattice": lat.tolist()}
        else:
            rs = {
                "Atoms": atoms,
                "CoordinateType": "Cartesian",
                "Lattice": lat.tolist(),
            }
        RSList.append(rs)

    URSList = []  # DS does not seem to read this part of the information, it can be left empty

    # data structure refs to inputs/2.15/neb.json
    data = {
        "BarrierInfo": {
            "MaxForce": maxForces.tolist(),
            "ReactionCoordinate": reactionCoordinates.tolist(),
            "Tangent": [],
            "TotalEnergy": totalEnergies.tolist(),
        },
        "Force": {"Tangent": tangents.tolist()},
        "LoopInfo": IterDict,
        "RelaxedStructure": RSList,
        "UnrelaxedStructure": URSList,
    }

    # ^ Write the dictionary to a json file
    import os

    absolute_output = os.path.abspath(output)
    with open(absolute_output, "w") as f:
        from json import dump

        dump(data, f, indent=4)

    print(f"==> {absolute_output}")


def _getef(directory: str = "."):
    """Read the energies and forces of each configuration during NEB calculation, NEB calculation can be non-converged
    But if the initial and final states are converged elsewhere, please manually move them into the 00 folders!

    Parameters
    ----------
    directory:
        Path for NEB calculation, default to current path

    Returns
    -------
    subfolders:
        List of configuration folder names
    resort_mfs:
        List of maximum components of configuration forces
    rcs:
        Reaction coordinate list
    ens:
        List of total electronic energies
    dEs:
        List of energy differences from the initial configuration

    """
    import os

    import numpy as np

    absdir = os.path.abspath(directory)
    subfolders = get_neb_subfolders(absdir)
    Nimage = len(subfolders)

    ens = []
    dEs = np.zeros(Nimage)
    rcs = [0]
    mfs = []

    # read energies
    import h5py

    count = 1
    for i, subfolder in enumerate(subfolders):
        if i == 0 or i == Nimage - 1:
            jsf = os.path.join(absdir, subfolder, f"system{subfolder}.json")
            old_jsf = os.path.join(absdir, subfolder, "system.json")
            hf = os.path.join(absdir, subfolder, "scf.h5")

            if os.path.isfile(hf):  # Prefer reading the content of the h5 file
                data = h5py.File(hf)
                en = np.asarray(data.get("/Energy/TotalEnergy0"))[0]
                if i == 0 or i == Nimage - 1:
                    mf = np.max(np.abs(np.asarray(data.get("/Force/ForceOnAtoms"))))
                    mfs.append(mf)

            elif os.path.isfile(jsf):  # Next, read the content of the JSON file
                with open(jsf) as f:
                    from json import load

                    data = load(f)
                en = data["Energy"]["TotalEnergy0"]
                if i == 0 or i == Nimage - 1:
                    mf = np.max(np.abs(data["Force"]["ForceOnAtoms"]))
                    mfs.append(mf)

            elif os.path.isfile(old_jsf):  # Compatible with old JSON
                with open(old_jsf) as f:
                    from json import load

                    data = load(f)
                en = data["Energy"]["TotalEnergy0"]
                if i == 0 or i == Nimage - 1:
                    mf = np.max(np.abs(data["Force"]["ForceOnAtoms"]))
                    mfs.append(mf)

            else:
                raise FileNotFoundError(f"No {jsf}/{old_jsf}/{hf} for {subfolder}")
            ens.append(en)

        else:
            jsf = os.path.join(absdir, subfolder, f"neb{subfolder}.json")
            sysjsf = os.path.join(absdir, subfolder, f"system{subfolder}.json")
            old_sysjsf = os.path.join(absdir, subfolder, "system.json")
            hf = os.path.join(absdir, subfolder, f"neb{subfolder}.h5")

            if os.path.isfile(hf):  # Prefer reading the content of the h5 file
                data = h5py.File(hf)
                en = np.asarray(data.get("/Energy/TotalEnergy0"))[0]
                mf = np.asarray(data.get("/MaxForce"))[-1]
                # the key may change depends on your DS-PAW version
                if "/Distance/Previous" in data:
                    rc = np.asarray(data.get("/Distance/Previous"))[-1]
                elif "/ReactionCoordinate" in data:
                    rc = np.asarray(data.get("/ReactionCoordinate"))[-2]
                else:
                    raise KeyError(
                        f"Neither /Distance/Previous nor /ReactionCoordinate in {hf}",
                    )
                rcs.append(rc)
                if count == Nimage - 2:  # before final image
                    if "/Distance/Next" in data:
                        rc = np.asarray(data.get("/Distance/Next"))[-1]
                    elif "/ReactionCoordinate" in data:
                        rc = np.asarray(data.get("/ReactionCoordinate"))[-1]
                    else:
                        raise KeyError(
                            f"Neither /Distance/Next nor /ReactionCoordinate in {hf}",
                        )
                    rcs.append(rc)

            elif os.path.isfile(jsf):
                from json import load

                if os.path.isfile(sysjsf):
                    with open(sysjsf) as f:
                        data = load(f)
                    en = data["Energy"]["TotalEnergy0"]
                elif os.path.isfile(old_sysjsf):  # Compatible with old DS-PAW
                    with open(old_sysjsf) as f:
                        data = load(f)
                    en = data["Energy"]["TotalEnergy0"]
                else:
                    raise FileNotFoundError(f"No {sysjsf}/{old_sysjsf}")

                with open(jsf) as f:
                    data = load(f)
                Nion_step = len(data)
                # en = data[Nion_step - 1]["TotalEnergy"] # invalid
                mf = data[Nion_step - 1]["MaxForce"]  # Maximum force at the last step
                rc = data[Nion_step - 1]["ReactionCoordinate"][
                    0
                ]  # Reaction coordinate of the final step
                rcs.append(rc)
                if count == Nimage - 2:  # before final image
                    rc = data[Nion_step - 1]["ReactionCoordinate"][
                        1
                    ]  # Final reaction coordinate
                    rcs.append(rc)

            else:
                raise FileNotFoundError(f"No {hf}/{jsf}")

            ens.append(en)
            mfs.append(mf)

            # get dE
            dE = ens[count] - ens[0]
            dEs[i] = dE
            count += 1
    dEs[-1] = ens[Nimage - 1] - ens[0]

    # The rcs read from nebXX.h5/json need to be converted to cumulative values
    for i in range(1, len(rcs)):
        rcs[i] += rcs[i - 1]

    rcs = np.asarray(rcs)

    return subfolders, mfs, rcs, ens, dEs


def monitor_force_energy(directory: str, outdir: str = ".", relative: bool = False):
    """Read forces and energies during NEB calculations from xx/DS-PAW.log and plot curves

    No JSON files are output during the calculation, and only force information is present in nebXX.h5 files, so DS-PAW.log must be read.

    Energy matching mode, should hit -40521.972259

    ------------------------------------------------------------
    LOOP 1:
    ------------------------------------------------------------
    #  iter |      Etot(eV)       dE(eV)         time
    #   1   |  -35958.655378  -3.595866e+04    47.784 s
    #   2   |  -40069.322436  -4.110667e+03    15.146 s
    #   3   |  -40490.281166  -4.209587e+02    15.114 s
    #   4   |  -40521.972259  -3.169109e+01    17.936 s

    Examples
    --------
    >>> from dspawpy.diffusion.nebtools import monitor_force_energy
    >>> monitor_force_energy(
    ...     directory="tests/supplement/neb_unfinished",
    ...     outdir="imgs"
    ... ) # doctest: +ELLIPSIS
    Max Force shape: (57, 4)
    ┌───────────┬───────────┬───────────┬───────────┐
    │ Folder 01 ┆ Folder 02 ┆ Folder 03 ┆ Folder 04 │
    ╞═══════════╪═══════════╪═══════════╪═══════════╡
    │ 23.775228 ┆ 71.547767 ┆ 72.641234 ┆ 24.147289 │
    │ 22.683711 ┆ 68.595607 ┆ 69.704747 ┆ 23.0549   │
    │ 5.624252  ┆ 20.071221 ┆ 20.049429 ┆ 5.567894  │
    │ 5.354774  ┆ 19.631643 ┆ 19.599093 ┆ 5.425462  │
    │ 3.188546  ┆ 9.840143  ┆ 9.748006  ┆ 2.943709  │
    │ …         ┆ …         ┆ …         ┆ …         │
    │ 0.293867  ┆ 0.812679  ┆ 0.920251  ┆ 0.573649  │
    │ 0.27249   ┆ 0.7475    ┆ 0.921836  ┆ 0.540239  │
    │ 0.299767  ┆ 0.360673  ┆ 1.174016  ┆ 0.416171  │
    │ 0.249903  ┆ 0.288985  ┆ 1.169237  ┆ 0.366117  │
    │ 0.204396  ┆ 0.518356  ┆ 0.913792  ┆ 0.300884  │
    └───────────┴───────────┴───────────┴───────────┘
    Energies shape: (57, 4)
    ┌───────────────┬───────────────┬───────────────┬───────────────┐
    │ Folder 01     ┆ Folder 02     ┆ Folder 03     ┆ Folder 04     │
    ╞═══════════════╪═══════════════╪═══════════════╪═══════════════╡
    │ -40448.281556 ┆ -40436.419243 ┆ -40436.084611 ┆ -40447.527434 │
    │ -40448.491374 ┆ -40437.026948 ┆ -40436.685178 ┆ -40447.73947  │
    │ -40451.391617 ┆ -40446.884408 ┆ -40446.613158 ┆ -40450.686918 │
    │ -40451.448662 ┆ -40447.079933 ┆ -40446.803281 ┆ -40450.743777 │
    │ -40452.126865 ┆ -40450.274376 ┆ -40449.978142 ┆ -40451.405157 │
    │ …             ┆ …             ┆ …             ┆ …             │
    │ -40452.620987 ┆ -40452.538682 ┆ -40452.230568 ┆ -40452.056262 │
    │ -40452.621777 ┆ -40452.544298 ┆ -40452.231776 ┆ -40452.055815 │
    │ -40452.620701 ┆ -40452.565649 ┆ -40452.164604 ┆ -40452.035357 │
    │ -40452.621371 ┆ -40452.569113 ┆ -40452.164784 ┆ -40452.037426 │
    │ -40452.622418 ┆ -40452.577864 ┆ -40452.141919 ┆ -40452.037885 │
    └───────────────┴───────────────┴───────────────┴───────────────┘
    ==> ...MaxForce.png...
    ==> ...Energies.png...

    """
    import os

    absdir = os.path.abspath(directory)
    absoutdir = os.path.abspath(outdir)
    subfolders = get_neb_subfolders(absdir)
    Nimage = len(subfolders)

    import re

    forces_all_images = []
    energies_all_images = []
    folder_names = []
    for i, subfolder in enumerate(subfolders):
        logfile = f"{absdir}/{subfolder}/DS-PAW.log"
        if i == 0 or i == Nimage - 1 or not os.path.isfile(logfile):
            continue

        folder_names.append(subfolder)
        forces = []
        if os.path.isfile(logfile):
            force_lines = []
            with open(logfile) as f:
                lines = f.readlines()
                # find all "LOOP 1:" lines
                inds = []
                for j, line in enumerate(lines):
                    if line.startswith("LOOP 1:"):
                        inds.append(j)
                if len(inds) == 0:
                    raise FileNotFoundError(f"No LOOP 1: in {logfile}")

                pattern = r"-{10,}\nLOOP \d+:\n-{10,}\n(?:.*\n)*?#\s*\d+\s+\|\s+(-\d+\.\d+).*\n\s*\n"
                latest_lines = lines[inds[-1] - 2 :]
                energies = re.findall(pattern, "".join(latest_lines))
                for line in latest_lines:
                    if "Max force:" in line:
                        force_lines.append(line)
                for line in force_lines:
                    forces.append(float(line.split()[2]))
            energies = [float(e) for e in energies]

            forces_all_images.append(forces)
            energies_all_images.append(energies)
        else:
            raise FileNotFoundError(f"No {logfile}")

    import polars as pl

    pl.Config.set_tbl_rows(10)

    # first check the LOOP number for each folder, and use the one all folders have as refs.
    nloops = [len(forces_all_images[i]) for i in range(len(folder_names))]
    if len(nloops) == 0:
        print("No loops found, did you specify the wrong path?")
        exit(1)
    min_loop = min(nloops)
    if min_loop == 0:
        print("please wait until at least the first loop is finished for all folders.")
        exit(1)
    max_loop = max(nloops)

    # set force and energy relative to last value of each folder
    if relative:
        for i in range(len(folder_names)):
            if len(forces_all_images[i]) == 0 or len(energies_all_images[i]) == 0:
                print(f"No forces or energies found in {folder_names[i]}/DS-PAW.log")
                exit(1)
            last_force = forces_all_images[i][min_loop - 1]
            last_energy = energies_all_images[i][min_loop - 1]
            forces_all_images[i] = [f - last_force for f in forces_all_images[i]]
            energies_all_images[i] = [e - last_energy for e in energies_all_images[i]]

    df1 = pl.DataFrame(
        {
            f"Folder {folder_names[i]}": forces_all_images[i]
            if len(forces_all_images[i]) == max_loop
            else forces_all_images[i] + [None] * (max_loop - len(forces_all_images[i]))
            for i in range(len(folder_names))
        },
    )
    print(f"Max Force {df1}")

    df2 = pl.DataFrame(
        {
            f"Folder {folder_names[i]}": energies_all_images[i]
            if len(energies_all_images[i]) == max_loop
            else energies_all_images[i]
            + [None] * (max_loop - len(energies_all_images[i]))
            for i in range(len(folder_names))
        },
    )
    print(f"Energies {df2}")

    import matplotlib.pyplot as plt

    plt.clf()
    for _i in range(len(folder_names)):
        plt.plot(
            range(len(forces_all_images[_i])),
            forces_all_images[_i],
            "-o",
            label=f"Image {folder_names[_i]}",
        )

    os.makedirs(absoutdir, exist_ok=True)
    plt.legend()
    plt.title("Max Force")
    plt.tight_layout()
    plt.savefig(f"{absoutdir}/MaxForce.png")
    print(f"==> {absoutdir}/MaxForce.png")

    plt.clf()
    for _i in range(len(folder_names)):
        plt.plot(
            range(len(energies_all_images[_i])),
            energies_all_images[_i],
            "-o",
            label=f"Image {folder_names[_i]}",
        )

    plt.legend()
    plt.title("Energies")
    plt.tight_layout()
    plt.savefig(f"{absoutdir}/Energies.png")
    print(f"==> {absoutdir}/Energies.png")



from dspawpy.io.read import load_h5
import os
import numpy as np
import polars as pl
from pathlib import Path
from typing import Union


def read_dos(
    dosfile: Union[str, Path],
    mode: int = 5,
    hide_index: bool = False,
    fmt: str = "8.3f",
) -> pl.DataFrame:
    """Construct a dataframe of dos data from an h5 or json file.

    Parameters
    ----------
    bandfile: str | Path
        band data file location, such as 'band.json'.
    mode: int
        which projection mode, only valid for pband data.
    hide_index: bool
        hide index column in dataframe or not.
    fmt: str
        control display decimal, such as '8.3f'

    Examples
    --------
    >>> from dspawpy.io.export import read_dos
    >>> read_dos(dosfile='tests/2.5/dos.h5')
    efermi=4.8711625000000005 eV
    shape: (401, 3)
    ┌───────┬──────────┬──────────┐
    │ index ┆ energy   ┆ dos      │
    ╞═══════╪══════════╪══════════╡
    │ 1     ┆   -4.830 ┆    0.690 │
    │ 2     ┆   -4.780 ┆    0.719 │
    │ 3     ┆   -4.730 ┆    0.757 │
    │ 4     ┆   -4.680 ┆    0.809 │
    │ 5     ┆   -4.630 ┆    0.950 │
    │ …     ┆ …        ┆ …        │
    │ 397   ┆   14.970 ┆    0.044 │
    │ 398   ┆   15.020 ┆    0.041 │
    │ 399   ┆   15.070 ┆    0.039 │
    │ 400   ┆   15.120 ┆    0.037 │
    │ 401   ┆   15.170 ┆    0.035 │
    └───────┴──────────┴──────────┘
    """
    absfile = os.path.abspath(dosfile)

    def _read_tdos_h5(dos):
        energies = np.asarray(dos["/DosInfo/DosEnergy"])

        if dos["/DosInfo/SpinType"][0] != "collinear":
            densities = {
                "energy": energies,
                "dos": np.asarray(dos["/DosInfo/Spin1/Dos"]),
            }
        else:
            densities = {
                "energy": energies,
                "up": np.asarray(dos["/DosInfo/Spin1/Dos"]),
                "down": np.asarray(dos["/DosInfo/Spin2/Dos"]),
            }
        return pl.DataFrame(data=densities)

    def _read_tdos_json(dos):
        energies = np.asarray(dos["DosInfo"]["DosEnergy"])

        if dos["DosInfo"]["SpinType"] != "collinear":
            densities = {
                "energy": energies,
                "dos": np.asarray(dos["DosInfo"]["Spin1"]["Dos"]),
            }
        else:
            densities = {
                "energy": energies,
                "up": np.asarray(dos["DosInfo"]["Spin1"]["Dos"]),
                "down": np.asarray(dos["DosInfo"]["Spin2"]["Dos"]),
            }
        return pl.DataFrame(data=densities)

    def _read_pdos_h5(dos, mode):
        energies: list[float] = dos["/DosInfo/DosEnergy"]
        data = {}
        orbitals: list[str] = dos["/DosInfo/Orbit"]

        atom_index: int = dos["/DosInfo/Spin1/ProjectDos/AtomIndexs"][0]  # 2
        orb_index: int = dos["/DosInfo/Spin1/ProjectDos/OrbitIndexs"][0]  # 9
        if dos["/DosInfo/SpinType"] == "collinear":
            for ai in range(atom_index):
                for oi in range(orb_index):
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}-up": dos[
                                f"/DosInfo/Spin1/ProjectDos{ai + 1}/{oi + 1}"
                            ]
                        }
                    )
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}-down": dos[
                                f"/DosInfo/Spin2/ProjectDos{ai + 1}/{oi + 1}"
                            ]
                        }
                    )
        else:
            for ai in range(atom_index):
                for oi in range(orb_index):
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}": dos[
                                f"/DosInfo/Spin1/ProjectDos{ai + 1}/{oi + 1}"
                            ]
                        }
                    )

        _data = _refactor_data(energies, data, mode)

        return pl.DataFrame(_data)

    def _read_pdos_json(dos, mode):
        energies: list[float] = dos["DosInfo"]["DosEnergy"]
        data = {}
        orbitals: list[str] = dos["DosInfo"]["Orbit"]

        if dos["DosInfo"]["SpinType"] == "collinear":
            project = dos["DosInfo"]["Spin1"]["ProjectDos"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}-up": contrib})
            project = dos["DosInfo"]["Spin2"]["ProjectDos"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}-down": contrib})
        else:
            project = dos["DosInfo"]["Spin1"]["ProjectDos"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}": contrib})

        _data = _refactor_data(energies, data, mode)

        return pl.DataFrame(_data)

    def _refactor_data(energies, data, mode):
        if mode == 1:  # spdf
            _data = {"energy": energies}
            for k, v in data.items():
                ao = k.split("-")[0]
                _, o = _split_atomIndex_orbital(ao)
                o_1st = o[0]
                if o_1st in _data:
                    _data[o_1st] += np.asarray(v)
                else:
                    _data[o_1st] = np.asarray(v)
        elif mode == 2:  # spxpy...
            _data = {"energy": energies}
            for k, v in data.items():
                ao = k.split("-")[0]
                _, o = _split_atomIndex_orbital(ao)
                if o in _data:
                    _data[o] += np.asarray(v)
                else:
                    _data[o] = np.asarray(v)
        elif mode == 3:  # element
            elements: list[str] = [atom["Element"] for atom in dos["AtomInfo"]["Atoms"]]
            _data = {"energy": energies}
            for k, v in data.items():
                ao = k.split("-")[0]
                a, _ = _split_atomIndex_orbital(ao)
                e = elements[a - 1]
                if e in _data:
                    _data[e] += np.asarray(v)
                else:
                    _data[e] = np.asarray(v)
        elif mode == 4:  # atom+spdf
            _data = {"energy": energies}
            for k, v in data.items():
                ao = k.split("-")[0]
                a, o = _split_atomIndex_orbital(ao)
                o_1st = o[0]
                if o_1st in _data:
                    _data[f"{a}{o_1st}"] += np.asarray(v)
                else:
                    _data[f"{a}{o_1st}"] = np.asarray(v)
        elif mode == 5:  # atom+spxpy...
            _data = data
        elif mode == 6:  # atom+t2g/eg
            _data = {"energy": energies}
            for k, v in data.items():
                ao = k.split("-")[0]
                a, o = _split_atomIndex_orbital(ao)
                if o in ["dxy", "dxz", "dyz"]:
                    if a in _data:
                        _data[f"{a}t2g"] += np.asarray(v)
                    else:
                        _data[f"{a}t2g"] = np.asarray(v)
                elif o in ["dz2", "dx2y2"]:
                    if a in _data:
                        _data[f"{a}eg"] += np.asarray(v)
                    else:
                        _data[f"{a}eg"] = np.asarray(v)
        else:
            print(f"{mode=} not supported yet")
            exit(1)

        return _data

    if absfile.endswith(".h5"):
        dos = load_h5(absfile)
        efermi = dos["/DosInfo/EFermi"][0]
        print(f"{efermi=} eV")
        if not dos["/DosInfo/Project"][0]:
            df = _read_tdos_h5(dos)

        else:
            df = _read_pdos_h5(dos, mode)

    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            dos = load(fin)
        if not dos["DosInfo"]["Project"]:
            df = _read_tdos_json(dos)
        else:
            df = _read_pdos_json(dos, mode)

    else:
        raise TypeError(f"{absfile} must be h5 or json file!")

    df = _format_float_columns_as_str_mapelements(df, fmt)

    if hide_index:
        return df
    else:
        return df.with_row_index(name="index", offset=1)


def read_band(
    bandfile: Union[str, Path],
    mode: int = 5,
    hide_index: bool = False,
    fmt: str = "8.3f",
):
    """Construct a dataframe of band structure data from an h5 or json file.

    Parameters
    ----------
    bandfile: str | Path
        band data file location, such as 'band.json'.
    mode: int
        which projection mode, only valid for pband data.
    hide_index: bool
        hide index column in dataframe or not.
    fmt: str
        control display decimal, such as '8.3f'

    Examples
    --------
    >>> from dspawpy.io.export import read_band
    >>> read_band(bandfile='tests/2.3/band.h5')
    efermi=5.1696 eV
    shape: (150, 16)
    ┌───────┬──────────┬──────────┬──────────┬───┬──────────┬──────────┬──────────┬──────────┐
    │ index ┆ kx       ┆ ky       ┆ kz       ┆ … ┆ 9        ┆ 10       ┆ 11       ┆ 12       │
    ╞═══════╪══════════╪══════════╪══════════╪═══╪══════════╪══════════╪══════════╪══════════╡
    │ 1     ┆    0.000 ┆    0.500 ┆    0.246 ┆ … ┆   -4.950 ┆    7.858 ┆   -6.766 ┆    8.395 │
    │ 2     ┆    0.000 ┆    0.172 ┆    0.246 ┆ … ┆   -0.491 ┆    9.558 ┆    3.944 ┆   12.139 │
    │ 3     ┆    0.000 ┆    0.672 ┆    0.491 ┆ … ┆    1.294 ┆   11.707 ┆    4.708 ┆   12.457 │
    │ 4     ┆    0.017 ┆    0.500 ┆    0.233 ┆ … ┆    3.604 ┆   12.532 ┆    4.708 ┆   12.457 │
    │ 5     ┆    0.000 ┆    0.181 ┆    0.233 ┆ … ┆    7.334 ┆   13.710 ┆    7.369 ┆   13.616 │
    │ …     ┆ …        ┆ …        ┆ …        ┆ … ┆ …        ┆ …        ┆ …        ┆ …        │
    │ 146   ┆    0.155 ┆    0.272 ┆    0.483 ┆ … ┆    3.067 ┆    8.304 ┆    0.255 ┆   12.534 │
    │ 147   ┆    0.655 ┆    0.543 ┆    0.483 ┆ … ┆    3.906 ┆   12.505 ┆    3.982 ┆   15.529 │
    │ 148   ┆    0.500 ┆    0.259 ┆    0.500 ┆ … ┆    4.713 ┆   12.505 ┆    3.982 ┆   15.942 │
    │ 149   ┆    0.164 ┆    0.259 ┆    0.500 ┆ … ┆    7.675 ┆   12.908 ┆    6.606 ┆   16.040 │
    │ 150   ┆    0.664 ┆    0.517 ┆    0.500 ┆ … ┆    7.700 ┆   16.401 ┆    8.395 ┆   17.259 │
    └───────┴──────────┴──────────┴──────────┴───┴──────────┴──────────┴──────────┴──────────┘
    """
    absfile = os.path.abspath(bandfile)

    def _read_tband_h5(band):
        kcoord = np.asarray(band["/BandInfo/CoordinatesOfKPoints"])  # 3,nkpt
        kx = kcoord[0, :]
        ky = kcoord[1, :]
        kz = kcoord[2, :]
        data = {
            "kx": kx,
            "ky": ky,
            "kz": kz,
        }
        bands = np.asarray(band["/BandInfo/Spin1/BandEnergies"])  # nband, nkpt

        if not band["/BandInfo/IsProject"][0]:
            for i in range(bands.shape[0]):
                data[f"{i + 1}"] = bands[i, :]
        else:
            for i in range(bands.shape[0]):
                data[f"{i + 1}-up"] = bands[i, :]
            bands = np.asarray(band["/BandInfo/Spin2/BandEnergies"])  # nband, nkpt
            for i in range(bands.shape[0]):
                data[f"{i + 1}-down"] = bands[i, :]

        return pl.DataFrame(data)

    def _read_tband_json(band):
        kc: list[float] = band["BandInfo"]["CoordinatesOfKPoints"]
        nkpt: int = band["BandInfo"]["NumberOfKpoints"]
        nband: int = band["BandInfo"]["NumberOfBand"]
        kcoord = np.array(kc).reshape(3, nkpt)
        kx = kcoord[0, :]
        ky = kcoord[1, :]
        kz = kcoord[2, :]
        data = {
            "kx": kx,
            "ky": ky,
            "kz": kz,
        }
        bands = np.asarray(band["BandInfo"]["Spin1"]["BandEnergies"]).reshape(
            nband, nkpt
        )
        if not band["BandInfo"]["IsProject"]:
            for i in range(bands.shape[0]):
                data[f"{i + 1}"] = bands[i, :]
        else:
            for i in range(bands.shape[0]):
                data[f"{i + 1}-up"] = bands[i, :]
            bands = np.asarray(band["BandInfo"]["Spin2"]["BandEnergies"])  # nband, nkpt
            for i in range(bands.shape[0]):
                data[f"{i + 1}-down"] = bands[i, :]

        return pl.DataFrame(data)

    def _read_pband_h5(band, mode):
        kc = band["/BandInfo/CoordinatesOfKPoints"]
        nkpt: int = band["/BandInfo/NumberOfKpoints"][0]
        nband: int = band["/BandInfo/NumberOfBand"][0]
        kcoord = np.array(kc).reshape(3, nkpt)
        kx = kcoord[0, :]
        ky = kcoord[1, :]
        kz = kcoord[2, :]
        data = {
            "kx": kx,
            "ky": ky,
            "kz": kz,
        }
        orbitals: list[str] = band["/BandInfo/Orbit"]
        atom_index = band["/BandInfo/Spin1/ProjectBand/AtomIndex"][0]
        orb_index = band["/BandInfo/Spin1/ProjectBand/OrbitIndexs"][0]

        if band["/BandInfo/SpinType"] == "collinear":
            for ai in range(atom_index):
                for oi in range(orb_index):
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}-up": band[
                                f"/BandInfo/Spin1/ProjectBand/{ai + 1}/{oi + 1}"
                            ]
                        }
                    )
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}-down": band[
                                f"/BandInfo/Spin2/ProjectBand/{ai + 1}/{oi + 1}"
                            ]
                        }
                    )
        else:
            for ai in range(atom_index):
                for oi in range(orb_index):
                    data.update(
                        {
                            f"{ai + 1}{orbitals[oi]}": band[
                                f"/BandInfo/Spin1/ProjectBand/1/{ai + 1}/{oi + 1}"
                            ]
                        }
                    )

        elements = band["/AtomInfo/Elements"]
        # elements: list[str] = [atom["Element"] for atom in band["/AtomInfo/Elements"]]
        _data = _refactor_data(data, nkpt, nband, elements, mode)

        return pl.DataFrame(_data)

    def _read_pband_json(band, mode):
        kc: list[float] = band["BandInfo"]["CoordinatesOfKPoints"]
        nkpt: int = band["BandInfo"]["NumberOfKpoints"]
        nband: int = band["BandInfo"]["NumberOfBand"]
        kcoord = np.array(kc).reshape(3, nkpt)
        kx = kcoord[0, :]
        ky = kcoord[1, :]
        kz = kcoord[2, :]
        data = {
            "kx": kx,
            "ky": ky,
            "kz": kz,
        }
        orbitals: list[str] = band["BandInfo"]["Orbit"]
        if band["BandInfo"]["SpinType"] == "collinear":
            project = band["BandInfo"]["Spin1"]["ProjectBand"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}-up": contrib})
            project = band["BandInfo"]["Spin2"]["ProjectBand"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}-down": contrib})
        else:
            project = band["BandInfo"]["Spin1"]["ProjectBand"]
            for p in project:
                atom_index = p["AtomIndex"]
                orb_index = p["OrbitIndex"] - 1
                contrib = p["Contribution"]
                data.update({f"{atom_index}{orbitals[orb_index]}": contrib})

        elements: list[str] = [atom["Element"] for atom in band["AtomInfo"]["Atoms"]]
        _data = _refactor_data(data, nkpt, nband, elements, mode)

        return pl.DataFrame(_data)

    def _refactor_data(data, nkpt, nband, elements, mode):
        _data = {}
        if mode == 1:  # ele
            for k, v in data.items():
                if k.startswith("k"):
                    _data[k] = v
                else:
                    cont = np.asarray(v).reshape(nband, nkpt)
                    ao = k.split("-")[0]
                    a, _ = _split_atomIndex_orbital(ao)
                    e = elements[a - 1]
                    for b in range(nband):
                        key = f"{b + 1}-{e}"
                        if key in _data:
                            _data[key] += np.asarray(cont[b])
                        else:
                            _data[key] = np.asarray(cont[b])
        elif mode == 2:  # ele+spdf
            for k, v in data.items():
                if k.startswith("k"):
                    _data[k] = v
                else:
                    cont = np.asarray(v).reshape(nband, nkpt)
                    ao = k.split("-")[0]
                    a, o = _split_atomIndex_orbital(ao)
                    o_1st = o[0]
                    e = elements[a - 1]
                    for b in range(nband):
                        key = f"{b + 1}-{e}-{o_1st}"
                        if key in _data:
                            _data[key] += np.asarray(cont[b])
                        else:
                            _data[key] = np.asarray(cont[b])
        elif mode == 3:  # ele+pxpy
            for k, v in data.items():
                if k.startswith("k"):
                    _data[k] = v
                else:
                    cont = np.asarray(v).reshape(nband, nkpt)
                    ao = k.split("-")[0]
                    a, o = _split_atomIndex_orbital(ao)
                    e = elements[a - 1]
                    for b in range(nband):
                        key = f"{b + 1}-{e}-{o}"
                        _data[key] = np.asarray(cont[b])
        elif mode == 4:  # atom+spdf
            for k, v in data.items():
                if k.startswith("k"):
                    _data[k] = v
                else:
                    cont = np.asarray(v).reshape(nband, nkpt)
                    ao = k.split("-")[0]
                    a, o = _split_atomIndex_orbital(ao)
                    o_1st = o[0]

                    for b in range(nband):
                        key = f"{b + 1}-{a}-{o_1st}"
                        if key in _data:
                            _data[key] += np.asarray(cont[b])
                        else:
                            _data[key] = np.asarray(cont[b])
        elif mode == 5:  # atom+spxpy...
            for k, v in data.items():
                if k.startswith("k"):
                    _data[k] = v
                else:
                    cont = np.asarray(v).reshape(nband, nkpt)
                    ao = k.split("-")[0]
                    a, o = _split_atomIndex_orbital(ao)
                    for b in range(nband):
                        _data[f"{b + 1}-{a}-{o}"] = np.asarray(cont[b])
        else:
            print(f"{mode=} not supported yet")
            exit(1)

        return _data

    if absfile.endswith(".h5"):
        band = load_h5(absfile)
        efermi = band["/BandInfo/EFermi"][0]
        print(f"{efermi=} eV")
        if not band["/BandInfo/IsProject"][0]:
            df = _read_tband_h5(band)
        else:
            df = _read_pband_h5(band, mode)

    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            band = load(fin)
        if not band["BandInfo"]["IsProject"]:
            df = _read_tband_json(band)
        else:
            df = _read_pband_json(band, mode)

    else:
        raise TypeError(f"{absfile} must be h5 or json file!")

    df = _format_float_columns_as_str_mapelements(df, fmt)

    if hide_index:
        return df
    else:
        return df.with_row_index(name="index", offset=1)


def _format_float_columns_as_str_mapelements(
    df: pl.DataFrame, fmt: str
) -> pl.DataFrame:
    if not isinstance(fmt, str):
        print(f"Warning: Format '{fmt}' is not a string. Skipping formatting.")
        return df

    try:
        return df.with_columns(
            pl.col(pl.Float64).map_elements(
                lambda x: f"{x:{fmt}}", return_dtype=pl.String
            )
        )
    except ValueError as e:
        print(
            f"Error applying format '{fmt}': {e}. This likely means the format string is invalid for a float value. Skipping formatting."
        )
        return df
    except Exception as e:
        print(f"An unexpected error occurred: {e}. Skipping formatting.")
        return df


def _split_atomIndex_orbital(s: str) -> tuple[int, str]:
    first_letter_index = -1
    for i, char in enumerate(s):
        if not char.isdigit():
            first_letter_index = i
            break

    if (
        first_letter_index == -1
    ):  # No letters found, assume the whole string is the atomIndex
        return int(s), ""
    else:
        atom_index_str = s[:first_letter_index]
        orbital_str = s[first_letter_index:]
        return int(atom_index_str), orbital_str



import os
import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_input
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse, Dselect


def s8_1(language: Literal["EN", "CN"]) -> dict:
    "Input file generation of intermediate structure"

    if auto_test_cli:
        from dspawpy.diffusion.neb import (
            NEB as cli_NEB,
        )
        from dspawpy.diffusion.neb import (
            write_neb_structures as cli_write_neb_structures,
        )
        from dspawpy.io.structure import read as cli_read

        for D in list_ds:
            if D["menu"] == 81:
                init_struct = cli_read(D["inits"])[0]
                final_struct = cli_read(D["fins"])[0]

                neb = cli_NEB(init_struct, final_struct, D["nmiddle"] + 2)
                pbc = True if D["pbc"] == "y" else False
                if D["method"] == "Linear":
                    structures = neb.linear_interpolate(pbc)
                else:
                    try:
                        structures = neb.idpp_interpolate(pbc=pbc)
                    except Exception:
                        logger.error(Dresponse[language][7])
                        structures = neb.linear_interpolate(pbc)
                        logger.warning(Dresponse[language][8])

                absdir = os.path.abspath(D["outd"])
                os.makedirs(os.path.dirname(absdir), exist_ok=True)
                cli_write_neb_structures(structures, fmt="as", path=absdir)
                logger.info(f"{Dresponse[language][9]} {D['outd']}")

                yn = get_input(Dselect[language][10], ["y", "n"])
                if yn.lower().startswith("y"):
                    from dspawpy.diffusion.nebtools import (
                        write_json_chain as cli_write_json_chain,
                    )
                    from dspawpy.diffusion.nebtools import (
                        write_xyz_chain as cli_write_xyz_chain,
                    )

                    cli_write_xyz_chain(preview=True, directory=absdir)
                    cli_write_json_chain(preview=True, directory=absdir)
        D = {}
    else:

        def imp():
            global NEB, write_neb_structures, read

            from dspawpy.diffusion.neb import NEB, write_neb_structures
            from dspawpy.io.structure import read

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 81
        D["inits"] = prompt(Dio[language]["inits"], completer=pc)
        D["fins"] = prompt(Dio[language]["fins"], completer=pc)
        D["nmiddle"] = int(input(Dparameter[language][8]))
        D["method"] = get_input(Dselect[language][9], ["IDPP", "Linear"])
        D["pbc"] = get_input(Dselect[language][25], ["y", "n"])
        pbc = True if D["pbc"] == "y" else False
        D["outd"] = prompt(Dio[language]["outd"], completer=pc)

        if D["outd"] == "":
            D["outd"] = "."

        import_thread.join()
        logger.info(Dresponse[language][16])
        init_struct = read(D["inits"])[0]
        final_struct = read(D["fins"])[0]

        neb = NEB(init_struct, final_struct, D["nmiddle"] + 2)
        if D["method"] == "Linear":
            structures = neb.linear_interpolate(pbc)
        else:
            try:
                structures = neb.idpp_interpolate(pbc=pbc)
            except Exception:
                logger.error(Dresponse[language][7])
                structures = neb.linear_interpolate(pbc)
                logger.warning(Dresponse[language][8])

        absdir = os.path.abspath(D["outd"])
        os.makedirs(os.path.dirname(absdir), exist_ok=True)
        write_neb_structures(structures, fmt="as", path=absdir)
        logger.info(f"{Dresponse[language][9]} {D['outd']}")

        yn = get_input(Dselect[language][10], ["y", "n"])
        if yn.lower().startswith("y"):
            from dspawpy.diffusion.nebtools import write_json_chain, write_xyz_chain

            write_xyz_chain(preview=True, directory=absdir)
            write_json_chain(preview=True, directory=absdir)

    return D


def s8_2(language: Literal["EN", "CN"]) -> dict:
    """Plot energy barrier diagram"""
    if auto_test_cli:
        from dspawpy.diffusion.nebtools import plot_barrier as cli_plot_barrier

        for D in list_ds:
            if D["menu"] == 82:
                if os.path.isdir(D["ind"]):
                    cli_plot_barrier(
                        directory=D["ind"], figname=D["figure"], show=False
                    )
                elif os.path.isfile(D["ind"]):
                    cli_plot_barrier(datafile=D["ind"], figname=D["figure"], show=False)
                else:
                    raise TypeError(D["inf"])
        D = {}
    else:

        def imp():
            global plot_barrier

            from dspawpy.diffusion.nebtools import plot_barrier

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 82
        D["ind"] = prompt(Dio[language]["neb"], completer=pc)
        D["figure"] = prompt(Dio[language]["figure"], completer=pc)
        logger.info(Dresponse[language][10])
        import_thread.join()
        logger.info(Dresponse[language][16])

        if os.path.isdir(D["ind"]):
            plot_barrier(directory=D["ind"], figname=D["figure"], show=False)
        elif os.path.isfile(D["ind"]):
            plot_barrier(datafile=D["ind"], figname=D["figure"], show=False)
        else:
            raise TypeError(D["inf"])
    return D


def s8_3(language: Literal["EN", "CN"]) -> dict:
    """Overview of transition state calculations"""

    if auto_test_cli:
        from dspawpy.diffusion.nebtools import summary as cli_summary

        for D in list_ds:
            if D["menu"] == 83:
                absdir = os.path.abspath(D["outd"])
                os.makedirs(os.path.dirname(absdir), exist_ok=True)
                cli_summary(
                    directory=D["ind"],
                    outdir=absdir,
                    figname=f"{absdir}/neb_summary.png",
                    show=False,
                )

        D = {}
    else:

        def imp():
            global summary

            from dspawpy.diffusion.nebtools import summary

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 82
        D["ind"] = prompt(Dio[language]["nebdir"], completer=pc)
        D["outd"] = prompt(Dio[language]["outd"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        assert os.path.isdir(D["ind"])
        absdir = os.path.abspath(D["outd"])
        os.makedirs(os.path.dirname(absdir), exist_ok=True)
        summary(
            directory=D["ind"],
            outdir=absdir,
            figname=f"{absdir}/neb_summary.png",
            show=False,
        )
    return D


def s8_4(language: Literal["EN", "CN"]) -> dict:
    """NEB chain visualization"""

    if auto_test_cli:
        from dspawpy.diffusion.nebtools import (
            write_json_chain as cli_write_json_chain,
        )
        from dspawpy.diffusion.nebtools import (
            write_xyz_chain as cli_write_xyz_chain,
        )

        for D in list_ds:
            if D["menu"] == 84:
                if D["yn"] == "y":
                    step = 0
                else:
                    step = int(input(Dselect[language][11]))
                cli_write_xyz_chain(False, D["ind"], step, D["dst"])
                cli_write_json_chain(False, D["ind"], step, D["dst"])
        D = {}
    else:

        def imp():
            global write_json_chain, write_xyz_chain
            from dspawpy.diffusion.nebtools import write_json_chain, write_xyz_chain

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 84
        D["yn"] = get_input(Dselect[language][10], ["y", "n"])
        D["ind"] = prompt(Dio[language]["nebdir"], completer=pc)
        if D["yn"] == "y":
            step = 0
        else:
            step = int(input(Dselect[language][11]))
        D["dst"] = prompt(Dio[language]["outd"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        write_xyz_chain(False, D["ind"], step, D["dst"])
        write_json_chain(False, D["ind"], step, D["dst"])
    return D


def s8_5(language: Literal["EN", "CN"]) -> dict:
    """Calculate the distance between configurations"""
    if auto_test_cli:
        from dspawpy.diffusion.nebtools import get_distance as cli_get_distance
        from dspawpy.io.structure import read as cli_read

        for D in list_ds:
            if D["menu"] == 85:
                s1 = cli_read(D["infile1"])[0]
                s2 = cli_read(D["infile2"])[0]
                result = cli_get_distance(
                    s1.frac_coords,
                    s2.frac_coords,
                    s1.lattice.matrix,
                    s2.lattice.matrix,
                )
                logger.info(str(result))
        D = {}
    else:

        def imp():
            global read, get_distance
            from dspawpy.diffusion.nebtools import get_distance
            from dspawpy.io.structure import read

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 85
        D["infile1"] = prompt(Dparameter[language][12], completer=pc)
        D["infile2"] = prompt(Dparameter[language][13], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])
        s1 = read(D["infile1"])[0]
        s2 = read(D["infile2"])[0]
        result = get_distance(
            s1.frac_coords,
            s2.frac_coords,
            s1.lattice.matrix,
            s2.lattice.matrix,
        )
        logger.info(str(result))
    return D


def s8_6(language: Literal["EN", "CN"]) -> dict:
    """Calculate neb continuation"""

    if auto_test_cli:
        from dspawpy.diffusion.nebtools import restart as cli_restart

        for D in list_ds:
            if D["menu"] == 86:
                cli_restart(D["ind"], D["outd"])
        D = {}
    else:

        def imp():
            global restart
            from dspawpy.diffusion.nebtools import restart

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 86
        D["ind"] = prompt(Dio[language]["neb"], completer=pc)
        D["outd"] = prompt(Dio[language]["outd"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])
        restart(D["ind"], D["outd"])
    return D


def s8_7(language: Literal["EN", "CN"]) -> dict:
    """Convergence trend plots of forces and energies during neb calculations"""
    if auto_test_cli:
        from dspawpy.diffusion.nebtools import (
            monitor_force_energy as cli_monitor_force_energy,
        )

        for D in list_ds:
            if D["menu"] == 87:
                cli_monitor_force_energy(D["ind"], D["fig_dir"], D["relative"])
        D = {}
    else:

        def imp():
            global monitor_force_energy
            from dspawpy.diffusion.nebtools import monitor_force_energy

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 87
        D["ind"] = prompt(Dio[language]["neb_can_be_unfinished"], completer=pc)
        D["relative"] = True if prompt(Dselect[language][22]).startswith("y") else False
        D["fig_dir"] = prompt(Dio[language]["fig_dir"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])
        monitor_force_energy(D["ind"], D["fig_dir"], D["relative"])

    return D



class NEB:
    def __init__(self, initial_structure, final_structure, nimages: int):
        self.nimages = nimages
        self.initial_structure = initial_structure
        self.final_structure = final_structure

    def linear_interpolate(self, pbc: bool = False):
        """Pbc defaults to False to respect user intention"""
        from dspawpy.diffusion.pathfinder import IDPPSolver

        return IDPPSolver.from_endpoints(
            endpoints=[self.initial_structure, self.final_structure],
            nimages=self.nimages - 2,
            sort_tol=0,  # Lock atom indices
            pbc=pbc,
        ).structures

    def idpp_interpolate(
        self,
        maxiter: int = 1000,
        tol: float = 1e-5,
        gtol: float = 1e-3,
        step_size: float = 0.05,
        max_disp: float = 0.05,
        spring_const: float = 5.0,
        pbc: bool = False,
    ):
        from dspawpy.diffusion.pathfinder import IDPPSolver

        return IDPPSolver.from_endpoints(
            endpoints=[self.initial_structure, self.final_structure],
            nimages=self.nimages - 2,
            sort_tol=0,  # Lock atom indices
            pbc=pbc,
        ).run(maxiter, tol, gtol, step_size, max_disp, spring_const)


def write_neb_structures(
    structures: list,
    coords_are_cartesian: bool = True,
    fmt: str = "as",
    path: str = ".",
    prefix="structure",
):
    r"""Interpolate and generate intermediate configuration files

    Parameters
    ----------
    structures:
        Structure list
    coords_are_cartesian:
        Is the coordinate Cartesian
    fmt:
        Structure file type, default to "as"
    path:
        Save path
    prefix:
        Filename prefix, default to "structure", which will generate files like structure00.as, structure01.as, ...

    Returns
    -------
    file
        Saves the configuration file

    Examples
    --------
    First, read the .as file to create a structure object

    >>> from dspawpy.io.structure import read
    >>> init_struct = read("tests/2.15/00/structure00.as")[0]
    >>> final_struct = read("tests/2.15/04/structure04.as")[0]

    Then, interpolate and generate intermediate structure files

    >>> from dspawpy.diffusion.neb import NEB,write_neb_structures
    >>> neb = NEB(init_struct,final_struct,8)
    >>> structures = neb.linear_interpolate()   # Linear interpolation

    Interpolated structures can be saved to the neb folder.

    >>> write_neb_structures(structures, path="tests/outputs/doctest/11neb_interpolate_structures") # doctest: +ELLIPSIS
    ==> .../structure00...as
    ==> .../structure01...as
    ==> .../structure02...as
    ==> .../structure03...as
    ==> .../structure04...as
    ==> .../structure05...as
    ==> .../structure06...as
    ==> .../structure07...as

    """
    if fmt is None:
        fmt = "as"
    import os

    absdir = os.path.abspath(path)
    N = len(str(len(structures)))
    if N <= 2:
        N = 2
    from dspawpy.io.structure import write

    for i, structure in enumerate(structures):
        path_name = str(i).zfill(N)
        os.makedirs(os.path.join(absdir, path_name), exist_ok=True)
        filename = os.path.join(absdir, path_name, "%s%s.%s" % (prefix, path_name, fmt))
        write(structure, filename, fmt, coords_are_cartesian)
